from typing import Dict

from fastapi import APIRouter, Depends, Query

from feast.api.registry.rest.codegen_utils import render_feature_view_code
from feast.api.registry.rest.rest_utils import (
    create_grpc_pagination_params,
    create_grpc_sorting_params,
    get_object_relationships,
    get_pagination_params,
    get_relationships_for_objects,
    get_sorting_params,
    grpc_call,
    paginate_and_sort,
    parse_tags,
)
from feast.registry_server import RegistryServer_pb2
from feast.type_map import _convert_value_type_str_to_value_type
from feast.types import from_value_type


def _extract_feature_view_from_any(any_feature_view: dict) -> dict:
    """Extract the specific feature view type and data from an AnyFeatureView object.

    Args:
        any_feature_view: Dictionary containing the AnyFeatureView data

    Returns:
        Dictionary with 'type' and feature view data, or empty dict if no valid type found
    """
    for key, value in any_feature_view.items():
        if value:
            return {"type": key, **value}

    return {}


def extract_feast_types_from_fields(fields):
    types = set()
    for field in fields:
        value_type_enum = _convert_value_type_str_to_value_type(
            field.get("valueType", "").upper()
        )
        feast_type = from_value_type(value_type_enum)
        dtype = (
            feast_type.__name__ if hasattr(feast_type, "__name__") else str(feast_type)
        )
        types.add(dtype)
    return list(types)


def get_feature_view_router(grpc_handler) -> APIRouter:
    router = APIRouter()

    @router.get("/feature_views/all")
    def list_feature_views_all(
        allow_cache: bool = Query(default=True),
        page: int = Query(1, ge=1),
        limit: int = Query(50, ge=1, le=100),
        sort_by: str = Query(None),
        sort_order: str = Query("asc"),
        include_relationships: bool = Query(
            False, description="Include relationships for each feature view"
        ),
    ):
        projects_resp = grpc_call(
            grpc_handler.ListProjects,
            RegistryServer_pb2.ListProjectsRequest(allow_cache=allow_cache),
        )
        projects = projects_resp.get("projects", [])
        all_feature_views = []
        for project in projects:
            project_name = project["spec"]["name"]
            req = RegistryServer_pb2.ListAllFeatureViewsRequest(
                project=project_name,
                allow_cache=allow_cache,
            )
            response = grpc_call(grpc_handler.ListAllFeatureViews, req)
            any_feature_views = response.get("featureViews", [])
            for any_feature_view in any_feature_views:
                feature_view = _extract_feature_view_from_any(any_feature_view)
                if feature_view:
                    feature_view["project"] = project_name
                    all_feature_views.append(feature_view)
        paged_feature_views, pagination = paginate_and_sort(
            all_feature_views, page, limit, sort_by, sort_order
        )
        result = {
            "featureViews": paged_feature_views,
            "pagination": pagination,
        }
        if include_relationships:
            relationships_map = {}
            for project in projects:
                project_name = project["spec"]["name"]
                # Filter feature views for this project
                project_feature_views = [
                    fv for fv in all_feature_views if fv["project"] == project_name
                ]
                rels = get_relationships_for_objects(
                    grpc_handler,
                    project_feature_views,
                    "featureView",
                    project_name,
                    allow_cache,
                )
                relationships_map.update(rels)
            result["relationships"] = relationships_map
        return result

    @router.get("/feature_views/{name}")
    def get_any_feature_view(
        name: str,
        project: str = Query(...),
        include_relationships: bool = Query(
            False, description="Include relationships for this feature view"
        ),
        allow_cache: bool = Query(True),
    ):
        req = RegistryServer_pb2.GetAnyFeatureViewRequest(
            name=name,
            project=project,
            allow_cache=allow_cache,
        )
        response = grpc_call(grpc_handler.GetAnyFeatureView, req)
        any_feature_view = response.get("anyFeatureView", {})

        result = _extract_feature_view_from_any(any_feature_view)

        if include_relationships:
            relationships = get_object_relationships(
                grpc_handler, "featureView", name, project, allow_cache
            )
            result["relationships"] = relationships

        if result and "spec" in result:
            spec = result["spec"]
            fv_type = result.get("type", "featureView")
            features = spec.get("features", [])
            if not isinstance(features, list):
                features = []
            if fv_type == "onDemandFeatureView":
                class_name = "OnDemandFeatureView"
                sources = spec.get("sources", {})
                if sources:
                    source_exprs = []
                    for k, v in sources.items():
                        var_name = v.get("name", k) if isinstance(v, dict) else str(v)
                        source_exprs.append(f'"{k}": {var_name}')
                    source_name = "{" + ", ".join(source_exprs) + "}"
                else:
                    source_name = "source_feature_view"
            elif fv_type == "streamFeatureView":
                class_name = "StreamFeatureView"
                stream_source = spec.get("streamSource", {})
                source_name = stream_source.get("name", "stream_source")
            else:
                class_name = "FeatureView"
                source_views = spec.get("source_views") or spec.get("sourceViews")
                if source_views and isinstance(source_views, list):
                    source_vars = [sv.get("name", "source_view") for sv in source_views]
                    source_name = "[" + ", ".join(source_vars) + "]"
                else:
                    source = spec.get("source")
                    if isinstance(source, dict) and source.get("name"):
                        source_name = source["name"]
                    else:
                        batch_source = spec.get("batchSource", {})
                        source_name = batch_source.get("name", "driver_stats_source")

            # Entities
            entities = spec.get("entities") or []
            entities_str = ", ".join(entities)

            # Feature schema
            schema_lines = []
            for field in features:
                value_type_enum = _convert_value_type_str_to_value_type(
                    field.get("valueType", "").upper()
                )
                feast_type = from_value_type(value_type_enum)
                dtype = getattr(feast_type, "__name__", str(feast_type))
                desc = field.get("description")
                desc_str = f', description="{desc}"' if desc else ""
                schema_lines.append(
                    f'        Field(name="{field["name"]}", dtype={dtype}{desc_str}),'
                )

            # Feast types
            feast_types = extract_feast_types_from_fields(features)

            # Tags
            tags = spec.get("tags", {})
            tags_str = f"tags={tags}," if tags else ""

            # TTL
            ttl = spec.get("ttl")
            ttl_str = "timedelta(days=1)"
            if ttl:
                if isinstance(ttl, int):
                    ttl_str = f"timedelta(seconds={ttl})"
                elif isinstance(ttl, str) and ttl.endswith("s") and ttl[:-1].isdigit():
                    ttl_str = f"timedelta(seconds={int(ttl[:-1])})"

            # Online
            online = spec.get("online", True)

            # Build context
            context = dict(
                class_name=class_name,
                name=spec.get("name", "example"),
                entities_str=entities_str,
                ttl_str=ttl_str,
                schema_lines=schema_lines,
                online=online,
                source_name=source_name,
                tags_str=tags_str,
                feast_types=feast_types,
            )

            result["featureDefinition"] = render_feature_view_code(context)

        return result

    @router.get("/feature_views")
    def list_all_feature_views(
        project: str = Query(...),
        allow_cache: bool = Query(default=True),
        include_relationships: bool = Query(
            False, description="Include relationships for each feature view"
        ),
        entity: str = Query(None, description="Filter feature views by entity name"),
        feature: str = Query(None, description="Filter feature views by feature name"),
        feature_service: str = Query(
            None, description="Filter feature views by feature service name"
        ),
        data_source: str = Query(
            None, description="Filter feature views by data source name"
        ),
        tags: Dict[str, str] = Depends(parse_tags),
        pagination_params: dict = Depends(get_pagination_params),
        sorting_params: dict = Depends(get_sorting_params),
    ):
        req = RegistryServer_pb2.ListAllFeatureViewsRequest(
            project=project,
            allow_cache=allow_cache,
            tags=tags,
            entity=entity,
            feature=feature,
            feature_service=feature_service,
            data_source=data_source,
            pagination=create_grpc_pagination_params(pagination_params),
            sorting=create_grpc_sorting_params(sorting_params),
        )
        response = grpc_call(grpc_handler.ListAllFeatureViews, req)
        any_feature_views = response.get("featureViews", [])

        # Extract the specific type of feature view from each AnyFeatureView
        feature_views = []
        for any_feature_view in any_feature_views:
            feature_view = _extract_feature_view_from_any(any_feature_view)
            if feature_view:
                feature_views.append(feature_view)

        result = {
            "featureViews": feature_views,
            "pagination": response.get("pagination", {}),
        }

        if include_relationships:
            relationships = get_relationships_for_objects(
                grpc_handler, feature_views, "featureView", project, allow_cache
            )
            result["relationships"] = relationships

        return result

    return router



import click
import yaml

from feast import utils
from feast.cli.cli_options import tagsOption
from feast.errors import FeastObjectNotFoundException
from feast.repo_operations import create_feature_store


@click.group(name="feature-services")
def feature_services_cmd():
    """
    Access feature services
    """
    pass


@feature_services_cmd.command("describe")
@click.argument("name", type=click.STRING)
@click.pass_context
def feature_service_describe(ctx: click.Context, name: str):
    """
    Describe a feature service
    """
    store = create_feature_store(ctx)

    try:
        feature_service = store.get_feature_service(name)
    except FeastObjectNotFoundException as e:
        print(e)
        exit(1)

    print(
        yaml.dump(
            yaml.safe_load(str(feature_service)),
            default_flow_style=False,
            sort_keys=False,
        )
    )


@feature_services_cmd.command(name="list")
@tagsOption
@click.pass_context
def feature_service_list(ctx: click.Context, tags: list[str]):
    """
    List all feature services
    """
    store = create_feature_store(ctx)
    feature_services = []
    tags_filter = utils.tags_list_to_dict(tags)
    for feature_service in store.list_feature_services(tags=tags_filter):
        feature_names = []
        for projection in feature_service.feature_view_projections:
            feature_names.extend(
                [f"{projection.name}:{feature.name}" for feature in projection.features]
            )
        feature_services.append([feature_service.name, ", ".join(feature_names)])

    from tabulate import tabulate

    print(tabulate(feature_services, headers=["NAME", "FEATURES"], tablefmt="plain"))



import pyarrow as pa

from feast.infra.compute_engines.dag.model import DAGFormat
from feast.infra.compute_engines.dag.value import DAGValue


class ArrowTableValue(DAGValue):
    def __init__(self, data: pa.Table):
        super().__init__(data, DAGFormat.ARROW)

    def __repr__(self):
        return f"ArrowTableValue(schema={self.data.schema}, rows={self.data.num_rows})"



import importlib
import json
import logging
from typing import TYPE_CHECKING, Any, List, Optional, Set

from colorama import Fore, Style
from fastapi import status as HttpStatusCode

if TYPE_CHECKING:
    from grpc import StatusCode as GrpcStatusCode

from feast.field import Field

logger = logging.getLogger(__name__)


class FeastError(Exception):
    pass

    def grpc_status_code(self) -> "GrpcStatusCode":
        from grpc import StatusCode as GrpcStatusCode

        return GrpcStatusCode.INTERNAL

    def http_status_code(self) -> int:
        return HttpStatusCode.HTTP_500_INTERNAL_SERVER_ERROR

    def __str__(self) -> str:
        if hasattr(self, "__overridden_message__"):
            return str(getattr(self, "__overridden_message__"))
        return super().__str__()

    def __repr__(self) -> str:
        if hasattr(self, "__overridden_message__"):
            return f"{type(self).__name__}('{getattr(self, '__overridden_message__')}')"
        return super().__repr__()

    def to_error_detail(self) -> str:
        """
        Returns a JSON representation of the error for serialization purposes.

        Returns:
          str: a string representation of a JSON document including `module`, `class` and `message` fields.
        """

        m = {
            "module": f"{type(self).__module__}",
            "class": f"{type(self).__name__}",
            "message": f"{str(self)}",
        }
        return json.dumps(m)

    @staticmethod
    def from_error_detail(detail: str) -> Optional["FeastError"]:
        try:
            m = json.loads(detail)
            if all(f in m for f in ["module", "class", "message"]):
                module_name = m["module"]
                class_name = m["class"]
                message = m["message"]
                module = importlib.import_module(module_name)
                class_reference = getattr(module, class_name)

                instance = class_reference.__new__(class_reference)
                setattr(instance, "__overridden_message__", message)
                return instance
        except Exception as e:
            logger.warning(f"Invalid error detail: {detail}: {e}")
        return None


class DataSourceNotFoundException(FeastError):
    def __init__(self, path):
        super().__init__(
            f"Unable to find table at '{path}'. Please check that table exists."
        )


class DataSourceNoNameException(FeastError):
    def __init__(self):
        super().__init__(
            "Unable to infer a name for this data source. Either table or name must be specified."
        )


class DataSourceRepeatNamesException(FeastError):
    def __init__(self, ds_name: str):
        super().__init__(
            f"Multiple data sources share the same case-insensitive name {ds_name}."
        )


class FeastObjectNotFoundException(FeastError):
    pass

    def grpc_status_code(self) -> "GrpcStatusCode":
        from grpc import StatusCode as GrpcStatusCode

        return GrpcStatusCode.NOT_FOUND

    def http_status_code(self) -> int:
        return HttpStatusCode.HTTP_404_NOT_FOUND


class EntityNotFoundException(FeastObjectNotFoundException):
    def __init__(self, name, project=None):
        if project:
            super().__init__(f"Entity {name} does not exist in project {project}")
        else:
            super().__init__(f"Entity {name} does not exist")


class FeatureServiceNotFoundException(FeastObjectNotFoundException):
    def __init__(self, name, project=None):
        if project:
            super().__init__(
                f"Feature service {name} does not exist in project {project}"
            )
        else:
            super().__init__(f"Feature service {name} does not exist")


class FeatureViewNotFoundException(FeastObjectNotFoundException):
    def __init__(self, name, project=None):
        if project:
            super().__init__(f"Feature view {name} does not exist in project {project}")
        else:
            super().__init__(f"Feature view {name} does not exist")


class OnDemandFeatureViewNotFoundException(FeastObjectNotFoundException):
    def __init__(self, name, project=None):
        if project:
            super().__init__(
                f"On demand feature view {name} does not exist in project {project}"
            )
        else:
            super().__init__(f"On demand feature view {name} does not exist")


class RequestDataNotFoundInEntityDfException(FeastObjectNotFoundException):
    def __init__(self, feature_name, feature_view_name):
        super().__init__(
            f"Feature {feature_name} not found in the entity dataframe, but required by feature view {feature_view_name}"
        )


class RequestDataNotFoundInEntityRowsException(FeastObjectNotFoundException):
    def __init__(self, feature_names):
        super().__init__(
            f"Required request data source features {feature_names} not found in the entity rows, but required by feature views"
        )


class DataSourceObjectNotFoundException(FeastObjectNotFoundException):
    def __init__(self, name, project=None):
        if project:
            super().__init__(f"Data source {name} does not exist in project {project}")
        else:
            super().__init__(f"Data source {name} does not exist")


class S3RegistryBucketNotExist(FeastObjectNotFoundException):
    def __init__(self, bucket):
        super().__init__(f"S3 bucket {bucket} for the Feast registry does not exist")


class S3RegistryBucketForbiddenAccess(FeastObjectNotFoundException):
    def __init__(self, bucket):
        super().__init__(f"S3 bucket {bucket} for the Feast registry can't be accessed")


class SavedDatasetNotFound(FeastObjectNotFoundException):
    def __init__(self, name: str, project: str):
        super().__init__(f"Saved dataset {name} does not exist in project {project}")


class ValidationReferenceNotFound(FeastObjectNotFoundException):
    def __init__(self, name: str, project: str):
        super().__init__(
            f"Validation reference {name} does not exist in project {project}"
        )


class FeastProviderLoginError(FeastError):
    """Error class that indicates a user has not authenticated with their provider."""


class FeastProviderNotImplementedError(FeastError):
    def __init__(self, provider_name):
        super().__init__(f"Provider '{provider_name}' is not implemented")


class FeastRegistryNotSetError(FeastError):
    def __init__(self):
        super().__init__("Registry is not set, but is required")


class FeastFeatureServerTypeInvalidError(FeastError):
    def __init__(self, feature_server_type: str):
        super().__init__(
            f"Feature server type was set to {feature_server_type}, but this type is invalid"
        )


class FeastRegistryTypeInvalidError(FeastError):
    def __init__(self, registry_type: str):
        super().__init__(
            f"Feature server type was set to {registry_type}, but this type is invalid"
        )


class FeastModuleImportError(FeastError):
    def __init__(self, module_name: str, class_name: str):
        super().__init__(
            f"Could not import module '{module_name}' while attempting to load class '{class_name}'"
        )


class FeastClassImportError(FeastError):
    def __init__(self, module_name: str, class_name: str):
        super().__init__(
            f"Could not import class '{class_name}' from module '{module_name}'"
        )


class FeastExtrasDependencyImportError(FeastError):
    def __init__(self, extras_type: str, nested_error: str):
        message = (
            nested_error
            + "\n"
            + f"You may need run {Style.BRIGHT + Fore.GREEN}pip install 'feast[{extras_type}]'{Style.RESET_ALL}"
        )
        super().__init__(message)


class FeastOfflineStoreUnsupportedDataSource(FeastError):
    def __init__(self, offline_store_name: str, data_source_name: str):
        super().__init__(
            f"Offline Store '{offline_store_name}' does not support data source '{data_source_name}'"
        )


class FeatureNameCollisionError(FeastError):
    def __init__(self, feature_refs_collisions: List[str], full_feature_names: bool):
        if full_feature_names:
            collisions = [ref.replace(":", "__") for ref in feature_refs_collisions]
            error_message = (
                "To resolve this collision, please ensure that the feature views or their own features "
                "have different names. If you're intentionally joining the same feature view twice on "
                "different sets of entities, please rename one of the feature views with '.with_name'."
            )
        else:
            collisions = [ref.split(":")[1] for ref in feature_refs_collisions]
            error_message = (
                "To resolve this collision, either use the full feature name by setting "
                "'full_feature_names=True', or ensure that the features in question have different names."
            )

        feature_names = ", ".join(set(collisions))
        super().__init__(
            f"Duplicate features named {feature_names} found.\n{error_message}"
        )


class SpecifiedFeaturesNotPresentError(FeastError):
    def __init__(
        self,
        specified_features: List[Field],
        inferred_features: List[Field],
        feature_view_name: str,
    ):
        super().__init__(
            f"Explicitly specified features {specified_features} not found in inferred list of features "
            f"{inferred_features} for '{feature_view_name}'"
        )


class SavedDatasetLocationAlreadyExists(FeastError):
    def __init__(self, location: str):
        super().__init__(f"Saved dataset location {location} already exists.")


class FeastOfflineStoreInvalidName(FeastError):
    def __init__(self, offline_store_class_name: str):
        super().__init__(
            f"Offline Store Class '{offline_store_class_name}' should end with the string `OfflineStore`.'"
        )


class FeastOnlineStoreInvalidName(FeastError):
    def __init__(self, online_store_class_name: str):
        super().__init__(
            f"Online Store Class '{online_store_class_name}' should end with the string `OnlineStore`.'"
        )


class FeastInvalidAuthConfigClass(FeastError):
    def __init__(self, auth_config_class_name: str):
        super().__init__(
            f"Auth Config Class '{auth_config_class_name}' should end with the string `AuthConfig`.'"
        )


class FeastInvalidBaseClass(FeastError):
    def __init__(self, class_name: str, class_type: str):
        super().__init__(
            f"Class '{class_name}' should have `{class_type}` as a base class."
        )


class FeastOnlineStoreUnsupportedDataSource(FeastError):
    def __init__(self, online_store_name: str, data_source_name: str):
        super().__init__(
            f"Online Store '{online_store_name}' does not support data source '{data_source_name}'"
        )


class FeastEntityDFMissingColumnsError(FeastError):
    def __init__(self, expected, missing):
        super().__init__(
            f"The entity dataframe you have provided must contain columns {expected}, "
            f"but {missing} were missing."
        )


class FeastJoinKeysDuringMaterialization(FeastError):
    def __init__(
        self, source: str, join_key_columns: Set[str], source_columns: Set[str]
    ):
        super().__init__(
            f"The DataFrame from {source} being materialized must have at least {join_key_columns} columns present, "
            f"but these were missing: {join_key_columns - source_columns} "
        )


class DockerDaemonNotRunning(FeastError):
    def __init__(self):
        super().__init__(
            "The Docker Python sdk cannot connect to the Docker daemon. Please make sure you have"
            "the docker daemon installed, and that it is running."
        )


class RegistryInferenceFailure(FeastError):
    def __init__(self, repo_obj_type: str, specific_issue: str):
        super().__init__(
            f"Inference to fill in missing information for {repo_obj_type} failed. {specific_issue}. "
            "Try filling the information explicitly."
        )


class BigQueryJobStillRunning(FeastError):
    def __init__(self, job_id):
        super().__init__(f"The BigQuery job with ID '{job_id}' is still running.")


class BigQueryJobCancelled(FeastError):
    def __init__(self, job_id):
        super().__init__(f"The BigQuery job with ID '{job_id}' was cancelled")


class RedshiftCredentialsError(FeastError):
    def __init__(self):
        super().__init__("Redshift API failed due to incorrect credentials")


class RedshiftQueryError(FeastError):
    def __init__(self, details):
        super().__init__(f"Redshift SQL Query failed to finish. Details: {details}")


class RedshiftTableNameTooLong(FeastError):
    def __init__(self, table_name: str):
        super().__init__(
            f"Redshift table names have a maximum length of 127 characters, but the table name {table_name} has length {len(table_name)} characters."
        )


class SnowflakeCredentialsError(FeastError):
    def __init__(self):
        super().__init__("Snowflake Connector failed due to incorrect credentials")


class SnowflakeQueryError(FeastError):
    def __init__(self, details):
        super().__init__(f"Snowflake SQL Query failed to finish. Details: {details}")


class EntityTimestampInferenceException(FeastError):
    def __init__(self, expected_column_name: str):
        super().__init__(
            f"Please provide an entity_df with a column named {expected_column_name} representing the time of events."
        )


class FeatureViewMissingDuringFeatureServiceInference(FeastError):
    def __init__(self, feature_view_name: str, feature_service_name: str):
        super().__init__(
            f"Missing {feature_view_name} feature view during inference for {feature_service_name} feature service."
        )


class InvalidEntityType(FeastError):
    def __init__(self, entity_type: type):
        super().__init__(
            f"The entity dataframe you have provided must be a Pandas DataFrame or a SQL query, "
            f"but we found: {entity_type} "
        )


class ConflictingFeatureViewNames(FeastError):
    # TODO: print file location of conflicting feature views
    def __init__(self, feature_view_name: str):
        super().__init__(
            f"The feature view name: {feature_view_name} refers to feature views of different types."
        )


class FeastInvalidInfraObjectType(FeastError):
    def __init__(self):
        super().__init__("Could not identify the type of the InfraObject.")


class SnowflakeIncompleteConfig(FeastError):
    def __init__(self, e: KeyError):
        super().__init__(f"{e} not defined in a config file or feature_store.yaml file")


class SnowflakeQueryUnknownError(FeastError):
    def __init__(self, query: str):
        super().__init__(f"Snowflake query failed: {query}")


class InvalidFeaturesParameterType(FeastError):
    def __init__(self, features: Any):
        super().__init__(
            f"Invalid `features` parameter type {type(features)}. Expected one of List[str] and FeatureService."
        )


class EntitySQLEmptyResults(FeastError):
    def __init__(self, entity_sql: str):
        super().__init__(
            f"No entity values found from the specified SQL query to generate the entity dataframe: {entity_sql}."
        )


class EntityDFNotDateTime(FeastError):
    def __init__(self):
        super().__init__(
            "The entity dataframe specified does not have the timestamp field as a datetime."
        )


class PushSourceNotFoundException(FeastError):
    def __init__(self, push_source_name: str):
        super().__init__(f"Unable to find push source '{push_source_name}'.")

    def http_status_code(self) -> int:
        return HttpStatusCode.HTTP_422_UNPROCESSABLE_ENTITY


class ReadOnlyRegistryException(FeastError):
    def __init__(self):
        super().__init__("Registry implementation is read-only.")


class DataFrameSerializationError(FeastError):
    def __init__(self, input: Any):
        if isinstance(input, dict):
            super().__init__(
                f"Failed to serialize the provided dictionary into a pandas DataFrame: {input.keys()}"
            )
        else:
            super().__init__(
                "Failed to serialize the provided input into a pandas DataFrame"
            )


class PermissionNotFoundException(FeastError):
    def __init__(self, name, project):
        super().__init__(f"Permission {name} does not exist in project {project}")


class PermissionObjectNotFoundException(FeastObjectNotFoundException):
    def __init__(self, name, project=None):
        if project:
            super().__init__(f"Permission {name} does not exist in project {project}")
        else:
            super().__init__(f"Permission {name} does not exist")


class ProjectNotFoundException(FeastError):
    def __init__(self, project):
        super().__init__(f"Project {project} does not exist in registry")


class ProjectObjectNotFoundException(FeastObjectNotFoundException):
    def __init__(self, name, project=None):
        super().__init__(f"Project {name} does not exist")


class ZeroRowsQueryResult(FeastError):
    def __init__(self, query: str):
        super().__init__(f"This query returned zero rows:\n{query}")


class ZeroColumnQueryResult(FeastError):
    def __init__(self, query: str):
        super().__init__(f"This query returned zero columns:\n{query}")


class FeastPermissionError(FeastError, PermissionError):
    def __init__(self, details: str):
        super().__init__(f"Permission error:\n{details}")

    def grpc_status_code(self) -> "GrpcStatusCode":
        from grpc import StatusCode as GrpcStatusCode

        return GrpcStatusCode.PERMISSION_DENIED

    def http_status_code(self) -> int:
        return HttpStatusCode.HTTP_403_FORBIDDEN



import logging
import os
import uuid
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

import dill
import fsspec
import numpy as np
import pandas as pd
import pyarrow as pa
import ray
import ray.data
from ray.data import Dataset
from ray.data.context import DatasetContext

from feast.data_source import DataSource
from feast.dataframe import DataFrameEngine, FeastDataFrame
from feast.errors import (
    RequestDataNotFoundInEntityDfException,
    SavedDatasetLocationAlreadyExists,
)
from feast.feature_logging import LoggingConfig, LoggingSource
from feast.feature_view import DUMMY_ENTITY_ID, DUMMY_ENTITY_VAL, FeatureView
from feast.feature_view_utils import resolve_feature_view_source_with_fallback
from feast.infra.offline_stores.file_source import (
    FileLoggingDestination,
    FileSource,
    SavedDatasetFileStorage,
)
from feast.infra.offline_stores.offline_store import (
    OfflineStore,
    RetrievalJob,
    RetrievalMetadata,
)
from feast.infra.offline_stores.offline_utils import (
    get_entity_df_timestamp_bounds,
    get_pyarrow_schema_from_batch_source,
    infer_event_timestamp_from_entity_df,
)
from feast.infra.ray_shared_utils import (
    _build_required_columns,
    apply_field_mapping,
    ensure_timestamp_compatibility,
    is_ray_data,
    normalize_timestamp_columns,
)
from feast.infra.registry.base_registry import BaseRegistry
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.repo_config import FeastConfigBaseModel, RepoConfig
from feast.saved_dataset import SavedDatasetStorage, ValidationReference
from feast.type_map import (
    convert_array_column,
    convert_scalar_column,
    feast_value_type_to_pandas_type,
    pa_to_feast_value_type,
)
from feast.utils import _get_column_names, make_df_tzaware, make_tzaware

logger = logging.getLogger(__name__)


def _get_data_schema_info(
    data: Union[pd.DataFrame, Dataset, Any],
) -> Tuple[Dict[str, Any], List[str]]:
    """
    Extract schema information from DataFrame or Dataset.
    Args:
        data: DataFrame or Ray Dataset
    Returns:
        Tuple of (dtypes_dict, column_names)
    """
    if is_ray_data(data):
        schema = data.schema()
        dtypes = {}
        for i, col in enumerate(schema.names):
            field_type = schema.field(i).type
            try:
                pa_type_str = str(field_type).lower()
                feast_value_type = pa_to_feast_value_type(pa_type_str)
                pandas_type_str = feast_value_type_to_pandas_type(feast_value_type)
                dtypes[col] = pd.api.types.pandas_dtype(pandas_type_str)
            except Exception:
                dtypes[col] = pd.api.types.pandas_dtype("object")
        columns = schema.names
    else:
        assert isinstance(data, pd.DataFrame)
        dtypes = data.dtypes.to_dict()
        columns = list(data.columns)
    return dtypes, columns


def _apply_to_data(
    data: Union[pd.DataFrame, Dataset, Any],
    process_func: Callable[[pd.DataFrame], pd.DataFrame],
    inplace: bool = False,
) -> Union[pd.DataFrame, Dataset, Any]:
    """
    Apply a processing function to DataFrame or Dataset.
    Args:
        data: DataFrame or Ray Dataset to process
        process_func: Function that takes a DataFrame and returns a processed DataFrame
        inplace: Whether to modify DataFrame in place (only applies to pandas)
    Returns:
        Processed DataFrame or Dataset
    """
    if is_ray_data(data):
        return data.map_batches(process_func, batch_format="pandas")
    else:
        assert isinstance(data, pd.DataFrame)
        if not inplace:
            data = data.copy()
        return process_func(data)


def _handle_empty_dataframe_case(
    join_key_columns: List[str],
    feature_name_columns: List[str],
    timestamp_columns: List[str],
) -> pd.DataFrame:
    """
    Handle empty DataFrame case by creating properly structured empty DataFrame.
    Args:
        join_key_columns: List of join key columns
        feature_name_columns: List of feature columns
        timestamp_columns: List of timestamp columns
    Returns:
        Empty DataFrame with proper structure and column types
    """
    empty_columns = _build_required_columns(
        join_key_columns, feature_name_columns, timestamp_columns
    )
    df = pd.DataFrame(columns=empty_columns)
    for col in timestamp_columns:
        if col in df.columns:
            df[col] = df[col].astype("datetime64[ns, UTC]")
    return df


def _safe_infer_event_timestamp_column(
    data: Union[pd.DataFrame, Dataset], fallback_column: str = "event_timestamp"
) -> str:
    """
    Safely infer the event timestamp column.
    Works with both pandas DataFrames and Ray Datasets.
    Args:
        data: DataFrame or Ray Dataset to analyze
        fallback_column: Default column name to use if inference fails
    Returns:
        Inferred or fallback timestamp column name
    """
    try:
        dtypes, _ = _get_data_schema_info(data)
        return infer_event_timestamp_from_entity_df(dtypes)
    except Exception as e:
        logger.debug(
            f"Timestamp column inference failed: {e}, using fallback: {fallback_column}"
        )
        return fallback_column


def _safe_get_entity_timestamp_bounds(
    data: Union[pd.DataFrame, Dataset, Any], timestamp_column: str
) -> Tuple[Optional[datetime], Optional[datetime]]:
    """
    Safely get entity timestamp bounds.
    Works with both pandas DataFrames and Ray Datasets.
    Args:
        data: DataFrame or Ray Dataset
        timestamp_column: Name of timestamp column
    Returns:
        Tuple of (min_timestamp, max_timestamp) or (None, None) if failed
    """
    try:
        if is_ray_data(data):
            min_ts = data.min(timestamp_column)
            max_ts = data.max(timestamp_column)
        else:
            if timestamp_column in data.columns:
                min_ts, max_ts = get_entity_df_timestamp_bounds(data, timestamp_column)
            else:
                return None, None
        if hasattr(min_ts, "to_pydatetime"):
            min_ts = min_ts.to_pydatetime()
        elif isinstance(min_ts, pd.Timestamp):
            min_ts = min_ts.to_pydatetime()
        if hasattr(max_ts, "to_pydatetime"):
            max_ts = max_ts.to_pydatetime()
        elif isinstance(max_ts, pd.Timestamp):
            max_ts = max_ts.to_pydatetime()
        return min_ts, max_ts
    except Exception as e:
        logger.debug(
            f"Timestamp bounds extraction failed: {e}, falling back to manual calculation"
        )
        try:
            if is_ray_data(data):

                def extract_bounds(batch: pd.DataFrame) -> pd.DataFrame:
                    if timestamp_column in batch.columns and not batch.empty:
                        timestamps = pd.to_datetime(batch[timestamp_column], utc=True)
                        return pd.DataFrame(
                            {"min_ts": [timestamps.min()], "max_ts": [timestamps.max()]}
                        )
                    return pd.DataFrame({"min_ts": [None], "max_ts": [None]})

                bounds_ds = data.map_batches(extract_bounds, batch_format="pandas")
                bounds_df = bounds_ds.to_pandas()

                if not bounds_df.empty:
                    min_ts = bounds_df["min_ts"].min()
                    max_ts = bounds_df["max_ts"].max()

                    if pd.notna(min_ts) and pd.notna(max_ts):
                        return min_ts.to_pydatetime(), max_ts.to_pydatetime()
            else:
                assert isinstance(data, pd.DataFrame)
                if timestamp_column in data.columns:
                    timestamps = pd.to_datetime(data[timestamp_column], utc=True)
                    return (
                        timestamps.min().to_pydatetime(),
                        timestamps.max().to_pydatetime(),
                    )
        except Exception:
            pass

        return None, None


def _safe_validate_schema(
    config: RepoConfig,
    data_source: DataSource,
    table_columns: List[str],
    operation_name: str = "operation",
) -> Optional[Tuple[pa.Schema, List[str]]]:
    """
    Safely validate schema using offline_utils with graceful fallback.
    Args:
        config: Repo configuration
        data_source: Data source to validate against
        table_columns: Actual table column names
        operation_name: Name of operation for logging
    Returns:
        Tuple of (expected_schema, expected_columns) or None if validation fails
    """
    try:
        expected_schema, expected_columns = get_pyarrow_schema_from_batch_source(
            config, data_source
        )
        if set(expected_columns) != set(table_columns):
            logger.warning(
                f"Schema mismatch in {operation_name}:\n"
                f"  Expected columns: {expected_columns}\n"
                f"  Actual columns: {table_columns}"
            )
            if set(expected_columns) == set(table_columns):
                logger.info(f"Columns match but order differs for {operation_name}")
                return expected_schema, expected_columns
        else:
            logger.debug(f"Schema validation passed for {operation_name}")
            return expected_schema, expected_columns

    except Exception as e:
        logger.warning(
            f"Schema validation skipped for {operation_name} due to error: {e}"
        )
        logger.debug("Schema validation error details:", exc_info=True)
    return None


def _convert_feature_column_types(
    data: Union[pd.DataFrame, Dataset], feature_views: List[FeatureView]
) -> Union[pd.DataFrame, Dataset]:
    """
    Convert feature columns to appropriate pandas types using Feast's type mapping utilities.
    Works with both pandas DataFrames and Ray Datasets.
    Args:
        data: DataFrame or Ray Dataset containing feature data
        feature_views: List of feature views with type information
    Returns:
        DataFrame or Dataset with properly converted feature column types
    """

    def convert_batch(batch: pd.DataFrame) -> pd.DataFrame:
        batch = batch.copy()

        for fv in feature_views:
            for feature in fv.features:
                feat_name = feature.name
                if feat_name not in batch.columns:
                    continue
                try:
                    value_type = feature.dtype.to_value_type()
                    if value_type.name.endswith("_LIST"):
                        batch[feat_name] = convert_array_column(
                            batch[feat_name], value_type
                        )
                    else:
                        target_pandas_type = feast_value_type_to_pandas_type(value_type)
                        batch[feat_name] = convert_scalar_column(
                            batch[feat_name], value_type, target_pandas_type
                        )
                except Exception as e:
                    logger.warning(
                        f"Failed to convert feature {feat_name} to proper type: {e}"
                    )
                    continue
        return batch

    return _apply_to_data(data, convert_batch)


class RayOfflineStoreConfig(FeastConfigBaseModel):
    """
    Configuration for the Ray Offline Store.

    For detailed configuration options and examples, see the documentation:
    https://docs.feast.dev/reference/offline-stores/ray
    """

    type: Literal[
        "feast.offline_stores.contrib.ray_offline_store.ray.RayOfflineStore", "ray"
    ] = "ray"
    storage_path: Optional[str] = None
    ray_address: Optional[str] = None

    # Optimization settings
    broadcast_join_threshold_mb: Optional[int] = 100
    enable_distributed_joins: Optional[bool] = True
    max_parallelism_multiplier: Optional[int] = 2
    target_partition_size_mb: Optional[int] = 64
    window_size_for_joins: Optional[str] = "1H"

    # Logging settings
    enable_ray_logging: Optional[bool] = False

    # Ray configuration for resource management (memory, CPU limits)
    ray_conf: Optional[Dict[str, Any]] = None


class RayResourceManager:
    """
    Manages Ray cluster resources for optimal performance.
    # See: https://docs.feast.dev/reference/offline-stores/ray#resource-management-and-testing
    """

    def __init__(self, config: Optional[RayOfflineStoreConfig] = None) -> None:
        """
        Initialize the resource manager with cluster resource information.
        """
        self.config = config or RayOfflineStoreConfig()
        self.cluster_resources = ray.cluster_resources()
        self.available_memory = self.cluster_resources.get("memory", 8 * 1024**3)
        self.available_cpus = int(self.cluster_resources.get("CPU", 4))
        self.num_nodes = len(ray.nodes()) if ray.is_initialized() else 1

    def configure_ray_context(self) -> None:
        """
        Configure Ray DatasetContext for optimal performance based on available resources.
        """
        ctx = DatasetContext.get_current()

        if self.available_memory > 32 * 1024**3:
            ctx.target_shuffle_buffer_size = 2 * 1024**3
            ctx.target_max_block_size = 512 * 1024**2
        else:
            ctx.target_shuffle_buffer_size = 512 * 1024**2
            ctx.target_max_block_size = 128 * 1024**2
        ctx.min_parallelism = self.available_cpus
        multiplier = (
            self.config.max_parallelism_multiplier
            if self.config.max_parallelism_multiplier is not None
            else 2
        )
        ctx.max_parallelism = self.available_cpus * multiplier
        ctx.shuffle_strategy = "sort"  # type: ignore
        ctx.enable_tensor_extension_casting = False

        if not getattr(self.config, "enable_ray_logging", False):
            ctx.enable_progress_bars = False
            if hasattr(ctx, "verbose_progress"):
                ctx.verbose_progress = False

        if getattr(self.config, "enable_ray_logging", False):
            logger.info(
                f"Configured Ray context: {self.available_cpus} CPUs, "
                f"{self.available_memory // 1024**3}GB memory, {self.num_nodes} nodes"
            )

    def estimate_optimal_partitions(self, dataset_size_bytes: int) -> int:
        """
        Estimate optimal number of partitions for a dataset based on size and resources.
        """
        target_partition_size = (self.config.target_partition_size_mb or 64) * 1024**2
        size_based_partitions = max(1, dataset_size_bytes // target_partition_size)
        max_partitions = self.available_cpus * (
            self.config.max_parallelism_multiplier or 2
        )
        return min(size_based_partitions, max_partitions)

    def should_use_broadcast_join(
        self, dataset_size_bytes: int, threshold_mb: Optional[int] = None
    ) -> bool:
        """
        Determine if dataset is small enough for broadcast join.
        """
        threshold = (
            threshold_mb
            if threshold_mb is not None
            else (self.config.broadcast_join_threshold_mb or 100)
        )
        return dataset_size_bytes <= threshold * 1024**2

    def estimate_processing_requirements(
        self, dataset_size_bytes: int, operation_type: str
    ) -> Dict[str, Any]:
        """
        Estimate resource requirements for different operations.
        """
        memory_multiplier = {
            "read": 1.2,  # 20% overhead for reading
            "join": 3.0,  # 3x for join operations
            "aggregate": 2.0,  # 2x for aggregations
            "shuffle": 2.5,  # 2.5x for shuffling
        }
        required_memory = dataset_size_bytes * memory_multiplier.get(
            operation_type, 2.0
        )
        return {
            "required_memory": required_memory,
            "optimal_partitions": self.estimate_optimal_partitions(dataset_size_bytes),
            "can_fit_in_memory": required_memory <= self.available_memory * 0.8,
            "should_broadcast": self.should_use_broadcast_join(dataset_size_bytes),
        }


class RayDataProcessor:
    """
    Optimized data processing with Ray for feature store operations.
    """

    def __init__(self, resource_manager: RayResourceManager) -> None:
        """
        Initialize the data processor with a resource manager.
        """
        self.resource_manager = resource_manager

    def optimize_dataset_for_join(self, ds: Dataset, join_keys: List[str]) -> Dataset:
        """
        Optimize dataset partitioning for join operations.
        """
        dataset_size = ds.size_bytes()
        optimal_partitions = self.resource_manager.estimate_optimal_partitions(
            dataset_size
        )
        if not join_keys:
            # For datasets without join keys, use simple repartitioning
            return ds.repartition(num_blocks=optimal_partitions)
        # For datasets with join keys, use shuffle for better distribution
        return ds.random_shuffle(num_blocks=optimal_partitions)

    def _manual_point_in_time_join(
        self,
        batch_df: pd.DataFrame,
        features_df: pd.DataFrame,
        join_keys: List[str],
        feature_join_keys: List[str],
        timestamp_field: str,
        requested_feats: List[str],
    ) -> pd.DataFrame:
        """
        Perform manual point-in-time join when merge_asof fails.

        This method handles cases where merge_asof cannot be used due to:
        - Entity mapping (different column names)
        - Complex multi-entity joins
        - Sorting issues with the data
        """
        result = batch_df.copy()
        for feat in requested_feats:
            is_list_feature = False
            if feat in features_df.columns:
                sample_values = features_df[feat].dropna()
                if not sample_values.empty:
                    sample_value = sample_values.iloc[0]
                    if isinstance(sample_value, (list, np.ndarray)):
                        is_list_feature = True
                    elif (
                        features_df[feat].dtype == object
                        and sample_values.apply(
                            lambda x: isinstance(x, (list, np.ndarray))
                        ).any()
                    ):
                        is_list_feature = True

            if is_list_feature:
                result[feat] = [[] for _ in range(len(result))]
            else:
                if feat in features_df.columns and pd.api.types.is_datetime64_any_dtype(
                    features_df[feat]
                ):
                    result[feat] = pd.Series(
                        [pd.NaT] * len(result), dtype="datetime64[ns, UTC]"
                    )
                else:
                    result[feat] = np.nan

        for _, entity_row in batch_df.iterrows():
            entity_matches = pd.Series(
                [True] * len(features_df), index=features_df.index
            )
            for entity_key, feature_key in zip(join_keys, feature_join_keys):
                if entity_key in entity_row and feature_key in features_df.columns:
                    entity_value = entity_row[entity_key]
                    feature_column = features_df[feature_key]
                    if pd.api.types.is_scalar(entity_value):
                        entity_matches &= feature_column == entity_value
                    else:
                        if hasattr(entity_value, "__len__") and len(entity_value) > 0:
                            entity_matches &= feature_column.isin(entity_value)
                        else:
                            entity_matches &= pd.Series(
                                [False] * len(features_df), index=features_df.index
                            )
            if not entity_matches.any():
                continue
            matching_features = features_df[entity_matches]
            entity_timestamp = entity_row[timestamp_field]
            if timestamp_field in matching_features.columns:
                time_matches = matching_features[timestamp_field] <= entity_timestamp
                matching_features = matching_features[time_matches]
            if matching_features.empty:
                continue

            if timestamp_field in matching_features.columns:
                matching_features = matching_features.sort_values(timestamp_field)
                latest_feature = matching_features.iloc[-1]
            else:
                latest_feature = matching_features.iloc[-1]

            entity_index = entity_row.name
            for feat in requested_feats:
                if feat in latest_feature:
                    feature_value = latest_feature[feat]
                    if pd.api.types.is_scalar(feature_value):
                        if pd.notna(feature_value):
                            result.loc[entity_index, feat] = feature_value
                    elif isinstance(feature_value, (list, tuple, np.ndarray)):
                        result.at[entity_index, feat] = feature_value
                    else:
                        try:
                            if pd.notna(feature_value):
                                result.at[entity_index, feat] = feature_value
                        except (ValueError, TypeError):
                            if feature_value is not None:
                                result.at[entity_index, feat] = feature_value

        return result

    def broadcast_join_features(
        self,
        entity_ds: Dataset,
        feature_df: pd.DataFrame,
        join_keys: List[str],
        timestamp_field: str,
        requested_feats: List[str],
        full_feature_names: bool = False,
        feature_view_name: Optional[str] = None,
        original_join_keys: Optional[List[str]] = None,
    ) -> Dataset:
        """Perform broadcast join for small feature datasets."""

        # Put feature data in Ray object store for efficient broadcasting
        feature_ref = ray.put(feature_df)

        def join_batch_with_features(batch: pd.DataFrame) -> pd.DataFrame:
            """Join a batch with broadcast feature data."""
            features = ray.get(feature_ref)

            enable_logging = getattr(
                self.resource_manager.config, "enable_ray_logging", False
            )
            if enable_logging:
                logger.info(
                    f"Processing feature view {feature_view_name} with join keys {join_keys}"
                )

            if original_join_keys:
                feature_join_keys = original_join_keys
                entity_join_keys = join_keys
            else:
                feature_join_keys = join_keys
                entity_join_keys = join_keys

            feature_cols = [timestamp_field] + feature_join_keys + requested_feats

            available_feature_cols = [
                col for col in feature_cols if col in features.columns
            ]

            if timestamp_field not in available_feature_cols:
                raise ValueError(
                    f"Timestamp field '{timestamp_field}' not found in features columns: {list(features.columns)}"
                )

            missing_feats = [
                feat for feat in requested_feats if feat not in features.columns
            ]
            if missing_feats:
                raise ValueError(
                    f"Requested features {missing_feats} not found in features columns: {list(features.columns)}"
                )

            features_filtered = features[available_feature_cols].copy()

            batch = normalize_timestamp_columns(batch, timestamp_field, inplace=True)
            features_filtered = normalize_timestamp_columns(
                features_filtered, timestamp_field, inplace=True
            )

            if not entity_join_keys:
                batch_sorted = batch.sort_values(timestamp_field).reset_index(drop=True)
                features_sorted = features_filtered.sort_values(
                    timestamp_field
                ).reset_index(drop=True)
                result = pd.merge_asof(
                    batch_sorted,
                    features_sorted,
                    on=timestamp_field,
                    direction="backward",
                )
            else:
                for key in entity_join_keys:
                    if key not in batch.columns:
                        batch[key] = np.nan
                for key in feature_join_keys:
                    if key not in features_filtered.columns:
                        features_filtered[key] = np.nan
                batch_clean = batch.dropna(
                    subset=entity_join_keys + [timestamp_field]
                ).copy()
                features_clean = features_filtered.dropna(
                    subset=feature_join_keys + [timestamp_field]
                ).copy()
                if batch_clean.empty or features_clean.empty:
                    return batch.head(0)
                if timestamp_field in batch_clean.columns:
                    batch_sorted = batch_clean.sort_values(
                        timestamp_field, ascending=True
                    ).reset_index(drop=True)
                else:
                    batch_sorted = batch_clean.reset_index(drop=True)

                right_sort_columns = []
                for key in feature_join_keys:
                    if key in features_clean.columns:
                        right_sort_columns.append(key)
                if timestamp_field in features_clean.columns:
                    right_sort_columns.append(timestamp_field)
                if right_sort_columns:
                    features_clean = features_clean.drop_duplicates(
                        subset=right_sort_columns, keep="last"
                    )
                    features_sorted = features_clean.sort_values(
                        right_sort_columns, ascending=True
                    ).reset_index(drop=True)
                else:
                    features_sorted = features_clean.reset_index(drop=True)

                if (
                    timestamp_field in features_sorted.columns
                    and len(features_sorted) > 1
                ):
                    if feature_join_keys:
                        grouped = features_sorted.groupby(feature_join_keys, sort=False)
                        for name, group in grouped:
                            if not group[timestamp_field].is_monotonic_increasing:
                                features_sorted = features_sorted.sort_values(
                                    feature_join_keys + [timestamp_field],
                                    ascending=True,
                                ).reset_index(drop=True)
                                break
                    else:
                        if not features_sorted[timestamp_field].is_monotonic_increasing:
                            features_sorted = features_sorted.sort_values(
                                timestamp_field, ascending=True
                            ).reset_index(drop=True)

                try:
                    if feature_join_keys:
                        batch_dedup_cols = [
                            k for k in entity_join_keys if k in batch_sorted.columns
                        ]
                        if timestamp_field in batch_sorted.columns:
                            batch_dedup_cols.append(timestamp_field)
                        if batch_dedup_cols:
                            batch_sorted = batch_sorted.drop_duplicates(
                                subset=batch_dedup_cols, keep="last"
                            )
                        feature_dedup_cols = [
                            k for k in feature_join_keys if k in features_sorted.columns
                        ]
                        if timestamp_field in features_sorted.columns:
                            feature_dedup_cols.append(timestamp_field)
                        if feature_dedup_cols:
                            features_sorted = features_sorted.drop_duplicates(
                                subset=feature_dedup_cols, keep="last"
                            )

                    if feature_join_keys:
                        if entity_join_keys == feature_join_keys:
                            result = pd.merge_asof(
                                batch_sorted,
                                features_sorted,
                                on=timestamp_field,
                                by=entity_join_keys,
                                direction="backward",
                                suffixes=("", "_right"),
                            )
                        else:
                            result = pd.merge_asof(
                                batch_sorted,
                                features_sorted,
                                on=timestamp_field,
                                left_by=entity_join_keys,
                                right_by=feature_join_keys,
                                direction="backward",
                                suffixes=("", "_right"),
                            )
                    else:
                        result = pd.merge_asof(
                            batch_sorted,
                            features_sorted,
                            on=timestamp_field,
                            direction="backward",
                            suffixes=("", "_right"),
                        )

                except Exception as e:
                    if enable_logging:
                        logger.warning(
                            f"merge_asof didn't work: {e}, implementing manual point-in-time join"
                        )
                    result = self._manual_point_in_time_join(
                        batch_clean,
                        features_clean,
                        entity_join_keys,
                        feature_join_keys,
                        timestamp_field,
                        requested_feats,
                    )
            if full_feature_names and feature_view_name:
                for feat in requested_feats:
                    if feat in result.columns:
                        new_name = f"{feature_view_name}__{feat}"
                        result[new_name] = result[feat]
                        result = result.drop(columns=[feat])

            return result

        return entity_ds.map_batches(join_batch_with_features, batch_format="pandas")

    def windowed_temporal_join(
        self,
        entity_ds: Dataset,
        feature_ds: Dataset,
        join_keys: List[str],
        timestamp_field: str,
        requested_feats: List[str],
        window_size: Optional[str] = None,
        full_feature_names: bool = False,
        feature_view_name: Optional[str] = None,
        original_join_keys: Optional[List[str]] = None,
    ) -> Dataset:
        """Perform windowed temporal join for large datasets."""

        window_size = window_size or (
            self.resource_manager.config.window_size_for_joins or "1H"
        )
        entity_optimized = self.optimize_dataset_for_join(entity_ds, join_keys)
        feature_optimized = self.optimize_dataset_for_join(feature_ds, join_keys)
        entity_windowed = self._add_time_windows_and_source_marker(
            entity_optimized, timestamp_field, "entity", window_size
        )
        feature_windowed = self._add_time_windows_and_source_marker(
            feature_optimized, timestamp_field, "feature", window_size
        )
        combined_ds = entity_windowed.union(feature_windowed)
        result_ds = combined_ds.map_batches(
            self._apply_windowed_point_in_time_logic,
            batch_format="pandas",
            fn_kwargs={
                "timestamp_field": timestamp_field,
                "join_keys": join_keys,
                "requested_feats": requested_feats,
                "full_feature_names": full_feature_names,
                "feature_view_name": feature_view_name,
                "original_join_keys": original_join_keys,
            },
        )

        return result_ds

    def _add_time_windows_and_source_marker(
        self, ds: Dataset, timestamp_field: str, source_marker: str, window_size: str
    ) -> Dataset:
        """Add time windows and source markers to dataset."""

        def add_window_and_source(batch: pd.DataFrame) -> pd.DataFrame:
            batch = batch.copy()
            if timestamp_field in batch.columns:
                batch["time_window"] = (
                    pd.to_datetime(batch[timestamp_field])
                    .dt.floor(window_size)
                    .astype("datetime64[ns, UTC]")
                )
            batch["_data_source"] = source_marker
            return batch

        return ds.map_batches(add_window_and_source, batch_format="pandas")

    def _apply_windowed_point_in_time_logic(
        self,
        batch: pd.DataFrame,
        timestamp_field: str,
        join_keys: List[str],
        requested_feats: List[str],
        full_feature_names: bool = False,
        feature_view_name: Optional[str] = None,
        original_join_keys: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        """Apply point-in-time correctness within time windows."""

        if len(batch) == 0:
            return pd.DataFrame()

        result_chunks = []
        group_keys = ["time_window"] + join_keys

        for group_values, group_data in batch.groupby(group_keys):
            entity_data = group_data[group_data["_data_source"] == "entity"].copy()
            feature_data = group_data[group_data["_data_source"] == "feature"].copy()
            if len(entity_data) > 0 and len(feature_data) > 0:
                entity_clean = entity_data.drop(columns=["time_window", "_data_source"])
                feature_clean = feature_data.drop(
                    columns=["time_window", "_data_source"]
                )
                if join_keys:
                    merged = pd.merge_asof(
                        entity_clean.sort_values(join_keys + [timestamp_field]),
                        feature_clean.sort_values(join_keys + [timestamp_field]),
                        on=timestamp_field,
                        by=join_keys,
                        direction="backward",
                    )
                else:
                    merged = pd.merge_asof(
                        entity_clean.sort_values(timestamp_field),
                        feature_clean.sort_values(timestamp_field),
                        on=timestamp_field,
                        direction="backward",
                    )

                result_chunks.append(merged)
            elif len(entity_data) > 0:
                entity_clean = entity_data.drop(columns=["time_window", "_data_source"])
                for feat in requested_feats:
                    if feat not in entity_clean.columns:
                        entity_clean[feat] = np.nan
                result_chunks.append(entity_clean)

        if result_chunks:
            result = pd.concat(result_chunks, ignore_index=True)
            if full_feature_names and feature_view_name:
                for feat in requested_feats:
                    if feat in result.columns:
                        new_name = f"{feature_view_name}__{feat}"
                        result[new_name] = result[feat]
                        result = result.drop(columns=[feat])

            return result
        else:
            return pd.DataFrame()


class RayRetrievalJob(RetrievalJob):
    def __init__(
        self,
        dataset_or_callable: Union[
            Dataset, pd.DataFrame, Callable[[], Union[Dataset, pd.DataFrame]]
        ],
        staging_location: Optional[str] = None,
        config: Optional[RayOfflineStoreConfig] = None,
    ):
        self._dataset_or_callable = dataset_or_callable
        self._staging_location = staging_location
        self._config = config or RayOfflineStoreConfig()
        self._cached_df: Optional[pd.DataFrame] = None
        self._cached_dataset: Optional[Dataset] = None
        self._metadata: Optional[RetrievalMetadata] = None
        self._full_feature_names: bool = False
        self._on_demand_feature_views: Optional[List[OnDemandFeatureView]] = None
        self._feature_refs: List[str] = []
        self._entity_df: Optional[pd.DataFrame] = None
        self._prefer_ray_datasets: bool = True

    def _create_metadata(self) -> RetrievalMetadata:
        """Create metadata from the entity DataFrame and feature references."""
        if self._entity_df is not None:
            timestamp_col = _safe_infer_event_timestamp_column(
                self._entity_df, "event_timestamp"
            )
            min_timestamp, max_timestamp = _safe_get_entity_timestamp_bounds(
                self._entity_df, timestamp_col
            )

            keys = [col for col in self._entity_df.columns if col != timestamp_col]
        else:
            try:
                result = self._resolve()
                if isinstance(result, Dataset):
                    timestamp_col = _safe_infer_event_timestamp_column(
                        result, "event_timestamp"
                    )
                    min_timestamp, max_timestamp = _safe_get_entity_timestamp_bounds(
                        result, timestamp_col
                    )
                    schema = result.schema()
                    keys = [col for col in schema.names if col != timestamp_col]
                else:
                    min_timestamp = None
                    max_timestamp = None
                    keys = []
            except Exception:
                min_timestamp = None
                max_timestamp = None
                keys = []

        return RetrievalMetadata(
            features=self._feature_refs,
            keys=keys,
            min_event_timestamp=min_timestamp,
            max_event_timestamp=max_timestamp,
        )

    def _set_metadata_info(
        self, feature_refs: List[str], entity_df: pd.DataFrame
    ) -> None:
        """Set the feature references and entity DataFrame for metadata creation."""
        self._feature_refs = feature_refs
        self._entity_df = entity_df

    def _resolve(self) -> Union[Dataset, pd.DataFrame]:
        if callable(self._dataset_or_callable):
            result = self._dataset_or_callable()
        else:
            result = self._dataset_or_callable
        return result

    def _get_ray_dataset(self) -> Dataset:
        """Get the result as a Ray Dataset, converting if necessary."""
        if self._cached_dataset is not None:
            return self._cached_dataset

        result = self._resolve()
        if isinstance(result, Dataset):
            self._cached_dataset = result
            return result
        elif isinstance(result, pd.DataFrame):
            self._cached_dataset = ray.data.from_pandas(result)
            return self._cached_dataset
        else:
            raise ValueError(f"Unsupported result type: {type(result)}")

    def to_df(
        self,
        validation_reference: Optional[ValidationReference] = None,
        timeout: Optional[int] = None,
    ) -> pd.DataFrame:
        if self._cached_df is not None and not self.on_demand_feature_views:
            df = self._cached_df
        else:
            if self.on_demand_feature_views:
                df = super().to_df(
                    validation_reference=validation_reference, timeout=timeout
                )
            else:
                if self._prefer_ray_datasets:
                    ray_ds = self._get_ray_dataset()
                    df = ray_ds.to_pandas()
                else:
                    result = self._resolve()
                    if isinstance(result, pd.DataFrame):
                        df = result
                    else:
                        df = result.to_pandas()
                self._cached_df = df

        if validation_reference:
            try:
                from feast.dqm.errors import ValidationFailed

                validation_result = validation_reference.profile.validate(df)
                if not validation_result.is_success:
                    raise ValidationFailed(validation_result)
            except ImportError:
                logger.warning("DQM profiler not available, skipping validation")
            except Exception as e:
                logger.error(f"Validation failed: {e}")
                raise ValueError(f"Data validation failed: {e}")
        return df

    def to_arrow(
        self,
        validation_reference: Optional[ValidationReference] = None,
        timeout: Optional[int] = None,
    ) -> pa.Table:
        if self.on_demand_feature_views:
            return super().to_arrow(
                validation_reference=validation_reference, timeout=timeout
            )

        if self._prefer_ray_datasets:
            try:
                ray_ds = self._get_ray_dataset()
                if hasattr(ray_ds, "to_arrow"):
                    return ray_ds.to_arrow()
                else:
                    df = ray_ds.to_pandas()
                    return pa.Table.from_pandas(df)
            except Exception:
                df = self.to_df(
                    validation_reference=validation_reference, timeout=timeout
                )
                return pa.Table.from_pandas(df)
        else:
            result = self._resolve()
            if isinstance(result, pd.DataFrame):
                return pa.Table.from_pandas(result)
            else:
                df = result.to_pandas()
                return pa.Table.from_pandas(df)

    def to_feast_df(
        self,
        validation_reference: Optional[ValidationReference] = None,
        timeout: Optional[int] = None,
    ) -> FeastDataFrame:
        """
        Return the result as a FeastDataFrame with Ray engine.

        This preserves Ray's lazy execution by wrapping the Ray Dataset directly.
        """
        # If we have on-demand feature views, fall back to base class Arrow implementation
        if self.on_demand_feature_views:
            return super().to_feast_df(validation_reference, timeout)

        # Get the Ray Dataset directly (maintains lazy execution)
        ray_ds = self._get_ray_dataset()

        return FeastDataFrame(
            data=ray_ds,
            engine=DataFrameEngine.RAY,
        )

    def to_remote_storage(self) -> list[str]:
        if not self._staging_location:
            raise ValueError("Staging location must be set for remote materialization.")
        try:
            ray_ds = self._get_ray_dataset()
            RayOfflineStore._ensure_ray_initialized()
            output_uri = os.path.join(self._staging_location, str(uuid.uuid4()))
            ray_ds.write_parquet(output_uri)
            return [output_uri]
        except Exception as e:
            raise RuntimeError(f"Failed to write to remote storage: {e}")

    @property
    def metadata(self) -> Optional[RetrievalMetadata]:
        """Return metadata information about retrieval."""
        if self._metadata is None:
            self._metadata = self._create_metadata()
        return self._metadata

    @property
    def full_feature_names(self) -> bool:
        return self._full_feature_names

    @property
    def on_demand_feature_views(self) -> List[OnDemandFeatureView]:
        return self._on_demand_feature_views or []

    def to_sql(self) -> str:
        raise NotImplementedError("SQL export not supported for Ray offline store")

    def _to_df_internal(self, timeout: Optional[int] = None) -> pd.DataFrame:
        if self._prefer_ray_datasets:
            ray_ds = self._get_ray_dataset()
            return ray_ds.to_pandas()
        else:
            return self._resolve().to_pandas()

    def _to_arrow_internal(self, timeout: Optional[int] = None) -> pa.Table:
        if self._prefer_ray_datasets:
            ray_ds = self._get_ray_dataset()
            try:
                if hasattr(ray_ds, "to_arrow"):
                    return ray_ds.to_arrow()
                else:
                    df = ray_ds.to_pandas()
                    return pa.Table.from_pandas(df)
            except Exception:
                df = ray_ds.to_pandas()
                return pa.Table.from_pandas(df)
        else:
            result = self._resolve()
            if isinstance(result, pd.DataFrame):
                return pa.Table.from_pandas(result)
            else:
                df = result.to_pandas()
                return pa.Table.from_pandas(df)

    def persist(
        self,
        storage: SavedDatasetStorage,
        allow_overwrite: Optional[bool] = False,
        timeout: Optional[int] = None,
    ) -> str:
        """Persist the dataset to storage using Ray operations."""

        if not isinstance(storage, SavedDatasetFileStorage):
            raise ValueError(
                f"Ray offline store only supports SavedDatasetFileStorage, got {type(storage)}"
            )
        destination_path = storage.file_options.uri
        if not destination_path.startswith(("s3://", "gs://", "hdfs://")):
            if not allow_overwrite and os.path.exists(destination_path):
                raise SavedDatasetLocationAlreadyExists(location=destination_path)
        try:
            ray_ds = self._get_ray_dataset()

            if not destination_path.startswith(("s3://", "gs://", "hdfs://")):
                os.makedirs(os.path.dirname(destination_path), exist_ok=True)

            ray_ds.write_parquet(destination_path)

            return destination_path
        except Exception as e:
            raise RuntimeError(f"Failed to persist dataset to {destination_path}: {e}")

    def materialize(self) -> None:
        """Materialize the Ray dataset to improve subsequent access performance."""
        try:
            ray_ds = self._get_ray_dataset()
            materialized_ds = ray_ds.materialize()
            self._cached_dataset = materialized_ds

            if getattr(self._config, "enable_ray_logging", False):
                logger.info("Ray dataset materialized successfully")
        except Exception as e:
            logger.warning(f"Failed to materialize Ray dataset: {e}")

    def schema(self) -> pa.Schema:
        """Get the schema of the dataset efficiently using Ray operations."""
        try:
            ray_ds = self._get_ray_dataset()
            return ray_ds.schema()
        except Exception:
            df = self.to_df()
            return pa.Table.from_pandas(df).schema


class RayOfflineStore(OfflineStore):
    def __init__(self) -> None:
        self._staging_location: Optional[str] = None
        self._ray_initialized: bool = False
        self._resource_manager: Optional[RayResourceManager] = None
        self._data_processor: Optional[RayDataProcessor] = None

    @staticmethod
    def _suppress_ray_logging() -> None:
        """Suppress Ray and Ray Data logging completely."""
        import warnings

        # Suppress Ray warnings
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="ray")
        warnings.filterwarnings("ignore", category=UserWarning, module="ray")

        # Set environment variables to suppress Ray output
        os.environ["RAY_DISABLE_IMPORT_WARNING"] = "1"
        os.environ["RAY_SUPPRESS_UNVERIFIED_TLS_WARNING"] = "1"
        os.environ["RAY_LOG_LEVEL"] = "ERROR"
        os.environ["RAY_DATA_LOG_LEVEL"] = "ERROR"
        os.environ["RAY_DISABLE_PROGRESS_BARS"] = "1"

        # Suppress all Ray-related loggers
        ray_loggers = [
            "ray",
            "ray.data",
            "ray.data.dataset",
            "ray.data.context",
            "ray.data._internal.streaming_executor",
            "ray.data._internal.execution",
            "ray.data._internal",
            "ray.tune",
            "ray.serve",
            "ray.util",
            "ray._private",
        ]
        for logger_name in ray_loggers:
            logging.getLogger(logger_name).setLevel(logging.ERROR)

        # Configure DatasetContext to disable progress bars
        try:
            from ray.data.context import DatasetContext

            ctx = DatasetContext.get_current()
            ctx.enable_progress_bars = False
            if hasattr(ctx, "verbose_progress"):
                ctx.verbose_progress = False
        except Exception:
            pass  # Ignore if Ray Data is not available

    @staticmethod
    def _ensure_ray_initialized(config: Optional[RepoConfig] = None) -> None:
        """Ensure Ray is initialized with proper configuration."""
        ray_config = None
        if config and hasattr(config, "offline_store"):
            ray_config = config.offline_store
            if isinstance(ray_config, RayOfflineStoreConfig):
                if not ray_config.enable_ray_logging:
                    RayOfflineStore._suppress_ray_logging()

        if not ray.is_initialized():
            ray_init_kwargs: Dict[str, Any] = {
                "ignore_reinit_error": True,
                "include_dashboard": False,
            }

            if (
                ray_config
                and isinstance(ray_config, RayOfflineStoreConfig)
                and not ray_config.enable_ray_logging
            ):
                ray_init_kwargs.update(
                    {
                        "log_to_driver": False,
                        "logging_level": "ERROR",
                    }
                )

            if config and hasattr(config, "offline_store"):
                if isinstance(ray_config, RayOfflineStoreConfig):
                    if ray_config.ray_address:
                        ray_init_kwargs["address"] = ray_config.ray_address
                    else:
                        ray_init_kwargs.update(
                            {
                                "_node_ip_address": os.getenv(
                                    "RAY_NODE_IP", "127.0.0.1"
                                ),
                                "num_cpus": os.cpu_count() or 4,
                            }
                        )

                    if ray_config.ray_conf:
                        ray_init_kwargs.update(ray_config.ray_conf)
                else:
                    pass  # Use default initialization

            ray.init(**ray_init_kwargs)

        ctx = DatasetContext.get_current()
        ctx.shuffle_strategy = "sort"  # type: ignore
        ctx.enable_tensor_extension_casting = False

        if (
            ray_config
            and isinstance(ray_config, RayOfflineStoreConfig)
            and not ray_config.enable_ray_logging
        ):
            RayOfflineStore._suppress_ray_logging()

        if ray.is_initialized():
            cluster_resources = ray.cluster_resources()
            if (
                not ray_config
                or not isinstance(ray_config, RayOfflineStoreConfig)
                or ray_config.enable_ray_logging
            ):
                logger.info(
                    f"Ray cluster initialized with {cluster_resources.get('CPU', 0)} CPUs, "
                    f"{cluster_resources.get('memory', 0) / (1024**3):.1f}GB memory"
                )

    def _init_ray(self, config: RepoConfig) -> None:
        ray_config = config.offline_store
        assert isinstance(ray_config, RayOfflineStoreConfig)
        RayOfflineStore._ensure_ray_initialized(config)

        if not ray_config.enable_ray_logging:
            RayOfflineStore._suppress_ray_logging()

        if self._resource_manager is None:
            self._resource_manager = RayResourceManager(ray_config)
            self._resource_manager.configure_ray_context()
        if self._data_processor is None:
            self._data_processor = RayDataProcessor(self._resource_manager)

    def _get_source_path(self, source: DataSource, config: RepoConfig) -> str:
        if not isinstance(source, FileSource):
            raise ValueError("RayOfflineStore currently only supports FileSource")
        repo_path = getattr(config, "repo_path", None)
        uri = FileSource.get_uri_for_file_path(repo_path, source.path)
        return uri

    def _optimize_dataset_for_operation(self, ds: Dataset, operation: str) -> Dataset:
        """Optimize dataset for specific operations."""
        if self._resource_manager is None:
            return ds

        dataset_size = ds.size_bytes()
        requirements = self._resource_manager.estimate_processing_requirements(
            dataset_size, operation
        )

        if requirements["can_fit_in_memory"]:
            ds = ds.materialize()

        optimal_partitions = requirements["optimal_partitions"]
        current_partitions = ds.num_blocks()

        if current_partitions != optimal_partitions:
            if getattr(self._resource_manager.config, "enable_ray_logging", False):
                logger.debug(
                    f"Repartitioning dataset from {current_partitions} to {optimal_partitions} blocks"
                )
            ds = ds.repartition(num_blocks=optimal_partitions)

        return ds

    @staticmethod
    def offline_write_batch(
        config: RepoConfig,
        feature_view: FeatureView,
        table: pa.Table,
        progress: Optional[Callable[[int], Any]] = None,
    ) -> None:
        """Write batch data using Ray operations with performance monitoring."""
        import time

        start_time = time.time()

        RayOfflineStore._ensure_ray_initialized(config)

        repo_path = getattr(config, "repo_path", None) or os.getcwd()
        ray_config = config.offline_store
        assert isinstance(ray_config, RayOfflineStoreConfig)

        if not ray_config.enable_ray_logging:
            RayOfflineStore._suppress_ray_logging()
        assert isinstance(feature_view.batch_source, FileSource)

        validation_result = _safe_validate_schema(
            config, feature_view.batch_source, table.column_names, "offline_write_batch"
        )

        if validation_result:
            expected_schema, expected_columns = validation_result
            if expected_columns != table.column_names and set(expected_columns) == set(
                table.column_names
            ):
                if getattr(ray_config, "enable_ray_logging", False):
                    logger.info("Reordering table columns to match expected schema")
                table = table.select(expected_columns)

        batch_source_path = feature_view.batch_source.file_options.uri
        feature_path = FileSource.get_uri_for_file_path(repo_path, batch_source_path)

        ds = ray.data.from_arrow(table)

        try:
            if feature_path.endswith(".parquet"):
                if os.path.exists(feature_path):
                    existing_ds = ray.data.read_parquet(feature_path)
                    combined_ds = existing_ds.union(ds)
                    combined_ds.write_parquet(feature_path)
                else:
                    ds.write_parquet(feature_path)
            else:
                os.makedirs(feature_path, exist_ok=True)
                ds.write_parquet(feature_path)

            if progress:
                progress(table.num_rows)

        except Exception:
            if getattr(ray_config, "enable_ray_logging", False):
                logger.info("Falling back to pandas-based writing")
            df = table.to_pandas()
            if feature_path.endswith(".parquet"):
                if os.path.exists(feature_path):
                    existing_df = pd.read_parquet(feature_path)
                    combined_df = pd.concat([existing_df, df], ignore_index=True)
                    combined_df.to_parquet(feature_path, index=False)
                else:
                    df.to_parquet(feature_path, index=False)
            else:
                os.makedirs(feature_path, exist_ok=True)
                ds_fallback = ray.data.from_pandas(df)
                ds_fallback.write_parquet(feature_path)

            if progress:
                progress(table.num_rows)

        duration = time.time() - start_time
        if getattr(ray_config, "enable_ray_logging", False):
            logger.info(
                f"Ray offline_write_batch performance: {table.num_rows} rows in {duration:.2f}s "
                f"({table.num_rows / duration:.0f} rows/s)"
            )

    def online_write_batch(
        self,
        config: RepoConfig,
        table: pa.Table,
        progress: Optional[Callable[[int], Any]] = None,
    ) -> None:
        """Ray offline store doesn't support online writes."""
        raise NotImplementedError("Ray offline store doesn't support online writes")

    @staticmethod
    def _process_filtered_batch(
        batch: pd.DataFrame,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_columns: List[str],
        timestamp_field_mapped: str,
    ) -> pd.DataFrame:
        batch = make_df_tzaware(batch)
        if batch.empty:
            return _handle_empty_dataframe_case(
                join_key_columns, feature_name_columns, timestamp_columns
            )
        all_required_columns = _build_required_columns(
            join_key_columns, feature_name_columns, timestamp_columns
        )
        if not join_key_columns:
            batch[DUMMY_ENTITY_ID] = DUMMY_ENTITY_VAL
        available_columns = [
            col for col in all_required_columns if col in batch.columns
        ]
        batch = batch[available_columns]
        if (
            "event_timestamp" not in batch.columns
            and timestamp_field_mapped != "event_timestamp"
        ):
            if timestamp_field_mapped in batch.columns:
                batch["event_timestamp"] = batch[timestamp_field_mapped]
        return batch

    @staticmethod
    def _load_and_filter_dataset(
        source_path: str,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        start_date: Optional[datetime],
        end_date: Optional[datetime],
    ) -> pd.DataFrame:
        try:
            field_mapping = getattr(data_source, "field_mapping", None)
            ds = RayOfflineStore._create_filtered_dataset(
                source_path, timestamp_field, start_date, end_date
            )
            df = ds.to_pandas()
            if field_mapping:
                df = df.rename(columns=field_mapping)
            timestamp_field_mapped = (
                field_mapping.get(timestamp_field, timestamp_field)
                if field_mapping
                else timestamp_field
            )
            created_timestamp_column_mapped = (
                field_mapping.get(created_timestamp_column, created_timestamp_column)
                if field_mapping and created_timestamp_column
                else created_timestamp_column
            )
            timestamp_columns = [timestamp_field_mapped]
            if created_timestamp_column_mapped:
                timestamp_columns.append(created_timestamp_column_mapped)
            df = normalize_timestamp_columns(df, timestamp_columns, inplace=True)
            df = RayOfflineStore._process_filtered_batch(
                df,
                join_key_columns,
                feature_name_columns,
                timestamp_columns,
                timestamp_field_mapped,
            )
            existing_timestamp_columns = [
                col for col in timestamp_columns if col in df.columns
            ]
            if existing_timestamp_columns:
                df = df.sort_values(existing_timestamp_columns, ascending=False)
            df = df.reset_index(drop=True)
            return df
        except Exception as e:
            raise RuntimeError(f"Failed to load data from {source_path}: {e}")

    @staticmethod
    def _load_and_filter_dataset_ray(
        source_path: str,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        start_date: Optional[datetime],
        end_date: Optional[datetime],
    ) -> Dataset:
        try:
            field_mapping = getattr(data_source, "field_mapping", None)
            ds = RayOfflineStore._create_filtered_dataset(
                source_path, timestamp_field, start_date, end_date
            )
            if field_mapping:
                ds = apply_field_mapping(ds, field_mapping)
            timestamp_field_mapped = (
                field_mapping.get(timestamp_field, timestamp_field)
                if field_mapping
                else timestamp_field
            )
            created_timestamp_column_mapped = (
                field_mapping.get(created_timestamp_column, created_timestamp_column)
                if field_mapping and created_timestamp_column
                else created_timestamp_column
            )
            timestamp_columns = [timestamp_field_mapped]
            if created_timestamp_column_mapped:
                timestamp_columns.append(created_timestamp_column_mapped)
            # Exclude __log_timestamp from normalization as it's used for time range filtering
            exclude_columns = (
                ["__log_timestamp"] if "__log_timestamp" in timestamp_columns else []
            )
            ds = normalize_timestamp_columns(
                ds, timestamp_columns, exclude_columns=exclude_columns
            )
            ds = ds.map_batches(
                lambda batch: RayOfflineStore._process_filtered_batch(
                    batch,
                    join_key_columns,
                    feature_name_columns,
                    timestamp_columns,
                    timestamp_field_mapped,
                ),
                batch_format="pandas",
            )
            timestamp_columns_existing = [
                col for col in timestamp_columns if col in ds.schema().names
            ]
            if timestamp_columns_existing:
                ds = ds.sort(timestamp_columns_existing, descending=True)

            return ds
        except Exception as e:
            raise RuntimeError(f"Failed to load data from {source_path}: {e}")

    @staticmethod
    def _pull_latest_processing_ray(
        ds: Dataset,
        join_key_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        field_mapping: Optional[Dict[str, str]] = None,
    ) -> Dataset:
        """
        Ray-native processing for pull_latest operations with deduplication.
        Args:
            ds: Ray Dataset to process
            join_key_columns: List of join key columns
            timestamp_field: Name of the timestamp field
            created_timestamp_column: Optional created timestamp column
            field_mapping: Optional field mapping dictionary
        Returns:
            Ray Dataset with latest records only
        """
        if not join_key_columns:
            return ds

        timestamp_field_mapped = (
            field_mapping.get(timestamp_field, timestamp_field)
            if field_mapping
            else timestamp_field
        )
        created_timestamp_column_mapped = (
            field_mapping.get(created_timestamp_column, created_timestamp_column)
            if field_mapping and created_timestamp_column
            else created_timestamp_column
        )

        timestamp_columns = [timestamp_field_mapped]
        if created_timestamp_column_mapped:
            timestamp_columns.append(created_timestamp_column_mapped)

        def deduplicate_batch(batch: pd.DataFrame) -> pd.DataFrame:
            if batch.empty:
                return batch

            existing_timestamp_columns = [
                col for col in timestamp_columns if col in batch.columns
            ]

            sort_columns = join_key_columns + existing_timestamp_columns
            if sort_columns:
                batch = batch.sort_values(
                    sort_columns,
                    ascending=[True] * len(join_key_columns)
                    + [False] * len(existing_timestamp_columns),
                )
                batch = batch.drop_duplicates(subset=join_key_columns, keep="first")

            return batch

        return ds.map_batches(deduplicate_batch, batch_format="pandas")

    @staticmethod
    def pull_latest_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        start_date: datetime,
        end_date: datetime,
    ) -> RetrievalJob:
        store = RayOfflineStore()
        store._init_ray(config)

        source_path = store._get_source_path(data_source, config)

        def _load_ray_dataset():
            ds = store._load_and_filter_dataset_ray(
                source_path,
                data_source,
                join_key_columns,
                feature_name_columns,
                timestamp_field,
                created_timestamp_column,
                start_date,
                end_date,
            )
            field_mapping = getattr(data_source, "field_mapping", None)
            ds = store._pull_latest_processing_ray(
                ds,
                join_key_columns,
                timestamp_field,
                created_timestamp_column,
                field_mapping,
            )

            return ds

        def _load_pandas_fallback():
            return store._load_and_filter_dataset(
                source_path,
                data_source,
                join_key_columns,
                feature_name_columns,
                timestamp_field,
                created_timestamp_column,
                start_date,
                end_date,
            )

        try:
            return RayRetrievalJob(
                _load_ray_dataset,
                staging_location=config.offline_store.storage_path,
                config=config.offline_store,
            )
        except Exception as e:
            logger.warning(f"Ray-native processing failed: {e}, falling back to pandas")
            return RayRetrievalJob(
                _load_pandas_fallback,
                staging_location=config.offline_store.storage_path,
                config=config.offline_store,
            )

    @staticmethod
    def pull_all_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
    ) -> RetrievalJob:
        store = RayOfflineStore()
        store._init_ray(config)

        source_path = store._get_source_path(data_source, config)

        fs, path_in_fs = fsspec.core.url_to_fs(source_path)
        if not fs.exists(path_in_fs):
            raise FileNotFoundError(f"Parquet path does not exist: {source_path}")

        def _load_ray_dataset():
            return store._load_and_filter_dataset_ray(
                source_path,
                data_source,
                join_key_columns,
                feature_name_columns,
                timestamp_field,
                created_timestamp_column,
                start_date,
                end_date,
            )

        def _load_pandas_fallback():
            return store._load_and_filter_dataset(
                source_path,
                data_source,
                join_key_columns,
                feature_name_columns,
                timestamp_field,
                created_timestamp_column,
                start_date,
                end_date,
            )

        try:
            return RayRetrievalJob(
                _load_ray_dataset,
                staging_location=config.offline_store.storage_path,
                config=config.offline_store,
            )
        except Exception as e:
            logger.warning(f"Ray-native processing failed: {e}, falling back to pandas")
            return RayRetrievalJob(
                _load_pandas_fallback,
                staging_location=config.offline_store.storage_path,
                config=config.offline_store,
            )

    @staticmethod
    def write_logged_features(
        config: RepoConfig,
        data: Union[pa.Table, Path],
        source: LoggingSource,
        logging_config: LoggingConfig,
        registry: BaseRegistry,
    ) -> None:
        RayOfflineStore._ensure_ray_initialized(config)

        ray_config = getattr(config, "offline_store", None)
        if (
            ray_config
            and isinstance(ray_config, RayOfflineStoreConfig)
            and not ray_config.enable_ray_logging
        ):
            RayOfflineStore._suppress_ray_logging()

        destination = logging_config.destination
        assert isinstance(destination, FileLoggingDestination), (
            f"Ray offline store only supports FileLoggingDestination for logging, "
            f"got {type(destination)}"
        )

        repo_path = getattr(config, "repo_path", None) or os.getcwd()
        absolute_path = FileSource.get_uri_for_file_path(repo_path, destination.path)

        try:
            if isinstance(data, Path):
                ds = ray.data.read_parquet(str(data))
            else:
                ds = ray.data.from_arrow(data)

                # Normalize feature timestamp precision to seconds to match test expectations during write
                # Note: Don't normalize __log_timestamp as it's used for time range filtering
                def normalize_timestamps(batch: pd.DataFrame) -> pd.DataFrame:
                    batch = batch.copy()
                    for col in batch.columns:
                        if (
                            pd.api.types.is_datetime64_any_dtype(batch[col])
                            and col != "__log_timestamp"
                        ):
                            batch[col] = batch[col].dt.floor("s")
                    return batch

                ds = ds.map_batches(normalize_timestamps, batch_format="pandas")
            ds = ds.materialize()
            filesystem, resolved_path = FileSource.create_filesystem_and_path(
                absolute_path, destination.s3_endpoint_override
            )
            path_obj = Path(resolved_path)
            if path_obj.suffix == ".parquet":
                path_obj = path_obj.with_suffix("")
            if not absolute_path.startswith(("s3://", "gs://")):
                path_obj.mkdir(parents=True, exist_ok=True)
            ds.write_parquet(str(path_obj))
        except Exception as e:
            raise RuntimeError(f"Failed to write logged features: {e}")

    @staticmethod
    def create_saved_dataset_destination(
        config: RepoConfig,
        name: str,
        path: Optional[str] = None,
    ) -> SavedDatasetStorage:
        """Create a saved dataset destination for Ray offline store."""

        if path is None:
            ray_config = config.offline_store
            assert isinstance(ray_config, RayOfflineStoreConfig)
            base_storage_path = ray_config.storage_path or "/tmp/ray-storage"
            path = f"{base_storage_path}/saved_datasets/{name}.parquet"

        return SavedDatasetFileStorage(path=path)

    @staticmethod
    def _create_filtered_dataset(
        source_path: str,
        timestamp_field: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
    ) -> Dataset:
        """Helper method to create a filtered dataset based on timestamp range."""
        ds = ray.data.read_parquet(source_path)

        try:
            col_names = ds.schema().names
            if timestamp_field not in col_names:
                raise ValueError(
                    f"Timestamp field '{timestamp_field}' not found in columns: {col_names}"
                )
        except Exception as e:
            raise ValueError(f"Failed to get dataset schema: {e}")

        def normalize(dt):
            return make_tzaware(dt) if dt and dt.tzinfo is None else dt

        start_date = normalize(start_date)
        end_date = normalize(end_date)

        try:
            if start_date and end_date:

                def filter_by_timestamp_range(batch):
                    return (batch[timestamp_field] >= start_date) & (
                        batch[timestamp_field] <= end_date
                    )

                ds = ds.filter(filter_by_timestamp_range)
            elif start_date:

                def filter_by_start_date(batch):
                    return batch[timestamp_field] >= start_date

                ds = ds.filter(filter_by_start_date)
            elif end_date:

                def filter_by_end_date(batch):
                    return batch[timestamp_field] <= end_date

                ds = ds.filter(filter_by_end_date)
        except Exception as e:
            raise RuntimeError(f"Failed to filter dataset by timestamp: {e}")

        return ds

    @staticmethod
    def get_historical_features(
        config: RepoConfig,
        feature_views: List[FeatureView],
        feature_refs: List[str],
        entity_df: Union[pd.DataFrame, str],
        registry: BaseRegistry,
        project: str,
        full_feature_names: bool = False,
    ) -> RetrievalJob:
        store = RayOfflineStore()
        store._init_ray(config)

        # Load entity_df as Ray dataset for distributed processing
        if isinstance(entity_df, str):
            entity_ds = ray.data.read_csv(entity_df)
            entity_df_sample = entity_ds.limit(1000).to_pandas()
        else:
            entity_ds = ray.data.from_pandas(entity_df)
            entity_df_sample = entity_df.copy()

        entity_ds = ensure_timestamp_compatibility(entity_ds, ["event_timestamp"])
        on_demand_feature_views = OnDemandFeatureView.get_requested_odfvs(
            feature_refs, project, registry
        )
        for odfv in on_demand_feature_views:
            odfv_request_data_schema = odfv.get_request_data_schema()
            for feature_name in odfv_request_data_schema.keys():
                if feature_name not in entity_df_sample.columns:
                    raise RequestDataNotFoundInEntityDfException(
                        feature_name=feature_name,
                        feature_view_name=odfv.name,
                    )

        odfv_names = {odfv.name for odfv in on_demand_feature_views}
        regular_feature_views = [
            fv for fv in feature_views if fv.name not in odfv_names
        ]
        global_field_mappings = {}
        for fv in regular_feature_views:
            mapping = getattr(fv.batch_source, "field_mapping", None)
            if mapping:
                for k, v in mapping.items():
                    global_field_mappings[v] = k

        if global_field_mappings:
            cols_to_rename = {
                v: k
                for k, v in global_field_mappings.items()
                if v in entity_df_sample.columns
            }
            if cols_to_rename:
                entity_ds = apply_field_mapping(entity_ds, cols_to_rename)

        result_ds = entity_ds
        for fv in regular_feature_views:
            fv_feature_refs = [
                ref
                for ref in feature_refs
                if ref.startswith(fv.projection.name_to_use() + ":")
            ]
            if not fv_feature_refs:
                continue

            entities = fv.entities or []
            entity_objs = [registry.get_entity(e, project) for e in entities]
            original_join_keys, _, timestamp_field, created_col = _get_column_names(
                fv, entity_objs
            )

            if fv.projection.join_key_map:
                join_keys = [
                    fv.projection.join_key_map.get(key, key)
                    for key in original_join_keys
                ]
            else:
                join_keys = original_join_keys

            requested_feats = [ref.split(":", 1)[1] for ref in fv_feature_refs]

            available_feature_names = [f.name for f in fv.features]
            missing_feats = [
                f for f in requested_feats if f not in available_feature_names
            ]
            if missing_feats:
                raise KeyError(
                    f"Requested features {missing_feats} not found in feature view '{fv.name}' "
                    f"(available: {available_feature_names})"
                )

            source_info = resolve_feature_view_source_with_fallback(
                fv, config, is_materialization=False
            )

            # Read from the resolved data source
            source_path = store._get_source_path(source_info.data_source, config)
            feature_ds = ray.data.read_parquet(source_path)
            logger.info(
                f"Reading feature view {fv.name}: {source_info.source_description}"
            )

            # Apply transformation if available
            if source_info.has_transformation and source_info.transformation_func:
                transformation_serialized = dill.dumps(source_info.transformation_func)

                def apply_transformation_with_serialized_func(
                    batch: pd.DataFrame,
                ) -> pd.DataFrame:
                    if batch.empty:
                        return batch
                    try:
                        logger.debug(
                            f"Applying transformation to batch with columns: {list(batch.columns)}"
                        )
                        transformation_func = dill.loads(transformation_serialized)
                        result = transformation_func(batch)
                        logger.debug(
                            f"Transformation result has columns: {list(result.columns)}"
                        )
                        return result
                    except Exception as e:
                        logger.error(f"Transformation failed for {fv.name}: {e}")
                        return batch

                feature_ds = feature_ds.map_batches(
                    apply_transformation_with_serialized_func, batch_format="pandas"
                )
                logger.info(f"Applied transformation to feature view {fv.name}")
            elif source_info.has_transformation:
                logger.warning(
                    f"Feature view {fv.name} marked as having transformation but no UDF found"
                )

            feature_size = feature_ds.size_bytes() or 0

            field_mapping = getattr(fv.batch_source, "field_mapping", None)
            if field_mapping:
                feature_ds = apply_field_mapping(feature_ds, field_mapping)
                join_keys = [field_mapping.get(k, k) for k in join_keys]
                timestamp_field = field_mapping.get(timestamp_field, timestamp_field)
                if created_col:
                    created_col = field_mapping.get(created_col, created_col)

            if (
                timestamp_field != "event_timestamp"
                and timestamp_field not in entity_df_sample.columns
                and "event_timestamp" in entity_df_sample.columns
            ):

                def add_timestamp_field(batch: pd.DataFrame) -> pd.DataFrame:
                    batch = batch.copy()
                    batch[timestamp_field] = batch["event_timestamp"]
                    return batch

                result_ds = result_ds.map_batches(
                    add_timestamp_field, batch_format="pandas"
                )
                result_ds = normalize_timestamp_columns(result_ds, timestamp_field)

            if store._resource_manager is None:
                raise ValueError("Resource manager not initialized")
            requirements = store._resource_manager.estimate_processing_requirements(
                feature_size, "join"
            )

            if requirements["should_broadcast"]:
                # Use broadcast join for small feature datasets
                if getattr(store._resource_manager.config, "enable_ray_logging", False):
                    logger.info(
                        f"Using broadcast join for {fv.name} (size: {feature_size // 1024**2}MB)"
                    )
                feature_df = feature_ds.to_pandas()
                feature_df = ensure_timestamp_compatibility(
                    feature_df, [timestamp_field]
                )

                if store._data_processor is None:
                    raise ValueError("Data processor not initialized")
                result_ds = store._data_processor.broadcast_join_features(
                    result_ds,
                    feature_df,
                    join_keys,
                    timestamp_field,
                    requested_feats,
                    full_feature_names,
                    fv.projection.name_to_use(),
                    original_join_keys if fv.projection.join_key_map else None,
                )
            else:
                # Use distributed windowed join for large feature datasets
                if getattr(store._resource_manager.config, "enable_ray_logging", False):
                    logger.info(
                        f"Using distributed join for {fv.name} (size: {feature_size // 1024**2}MB)"
                    )
                feature_ds = ensure_timestamp_compatibility(
                    feature_ds, [timestamp_field]
                )

                if store._data_processor is None:
                    raise ValueError("Data processor not initialized")
                result_ds = store._data_processor.windowed_temporal_join(
                    result_ds,
                    feature_ds,
                    join_keys,
                    timestamp_field,
                    requested_feats,
                    window_size=config.offline_store.window_size_for_joins,
                    full_feature_names=full_feature_names,
                    feature_view_name=fv.projection.name_to_use(),
                    original_join_keys=original_join_keys
                    if fv.projection.join_key_map
                    else None,
                )

        def finalize_result(batch: pd.DataFrame) -> pd.DataFrame:
            batch = batch.copy()

            existing_columns = set(batch.columns)
            for col in entity_df_sample.columns:
                if col not in existing_columns:
                    if len(batch) <= len(entity_df_sample):
                        batch[col] = entity_df_sample[col].iloc[: len(batch)].values
                    else:
                        repeated_values = np.tile(
                            entity_df_sample[col].values,
                            (len(batch) // len(entity_df_sample) + 1),
                        )
                        batch[col] = repeated_values[: len(batch)]

            if "event_timestamp" not in batch.columns:
                if "event_timestamp" in entity_df_sample.columns:
                    batch["event_timestamp"] = (
                        entity_df_sample["event_timestamp"].iloc[: len(batch)].values
                    )
                    batch = normalize_timestamp_columns(
                        batch, "event_timestamp", inplace=True
                    )
                elif timestamp_field in batch.columns:
                    batch["event_timestamp"] = batch[timestamp_field]

            return batch

        result_ds = result_ds.map_batches(finalize_result, batch_format="pandas")
        result_ds = _convert_feature_column_types(result_ds, regular_feature_views)

        storage_path = config.offline_store.storage_path
        if not storage_path:
            raise ValueError("Storage path must be set in config")

        job = RayRetrievalJob(
            result_ds, staging_location=storage_path, config=config.offline_store
        )
        job._full_feature_names = full_feature_names
        job._on_demand_feature_views = on_demand_feature_views
        job._feature_refs = feature_refs
        job._entity_df = entity_df_sample
        job._metadata = job._create_metadata()
        return job



import click
import yaml

from feast import utils
from feast.cli.cli_options import tagsOption
from feast.errors import FeastObjectNotFoundException
from feast.repo_operations import create_feature_store


@click.group(name="saved-datasets")
def saved_datasets_cmd():
    """
    [Experimental] Access saved datasets
    """
    pass


@saved_datasets_cmd.command("describe")
@click.argument("name", type=click.STRING)
@click.pass_context
def saved_datasets_describe(ctx: click.Context, name: str):
    """
    [Experimental] Describe a saved dataset
    """
    store = create_feature_store(ctx)

    try:
        saved_dataset = store.get_saved_dataset(name)
    except FeastObjectNotFoundException as e:
        print(e)
        exit(1)

    print(
        yaml.dump(
            yaml.safe_load(str(saved_dataset)),
            default_flow_style=False,
            sort_keys=False,
        )
    )


@saved_datasets_cmd.command(name="list")
@tagsOption
@click.pass_context
def saved_datasets_list(ctx: click.Context, tags: list[str]):
    """
    [Experimental] List all saved datasets
    """
    store = create_feature_store(ctx)
    table = []
    tags_filter = utils.tags_list_to_dict(tags)
    for saved_dataset in store.list_saved_datasets(tags=tags_filter):
        table.append([saved_dataset.name])

    from tabulate import tabulate

    print(tabulate(table, headers=["NAME"], tablefmt="plain"))



import base64
import logging
import warnings
from datetime import datetime
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple

import pytz
from couchbase.auth import PasswordAuthenticator
from couchbase.cluster import Cluster
from couchbase.exceptions import (
    CollectionAlreadyExistsException,
    DocumentNotFoundException,
    ScopeAlreadyExistsException,
)
from couchbase.options import ClusterOptions
from pydantic import StrictStr

from feast import Entity, FeatureView, RepoConfig
from feast.infra.key_encoding_utils import serialize_entity_key
from feast.infra.online_stores.online_store import OnlineStore
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import FeastConfigBaseModel

logger = logging.getLogger(__name__)
warnings.simplefilter("once", RuntimeWarning)


class CouchbaseOnlineStoreConfig(FeastConfigBaseModel):
    """
    Configuration for the Couchbase online store.
    """

    type: Literal["couchbase.online"] = "couchbase.online"

    connection_string: Optional[StrictStr] = None
    user: Optional[StrictStr] = None
    password: Optional[StrictStr] = None
    bucket_name: Optional[StrictStr] = None
    kv_port: Optional[int] = None


class CouchbaseOnlineStore(OnlineStore):
    """
    An online store implementation that uses Couchbase.
    """

    _cluster = None

    def _get_conn(self, config: RepoConfig, scope_name: str, collection_name: str):
        """
        Obtain a connection to the Couchbase cluster and get the specific scope and collection.
        """
        online_store_config = config.online_store
        assert isinstance(online_store_config, CouchbaseOnlineStoreConfig)

        if not self._cluster:
            self._cluster = Cluster(
                f"{online_store_config.connection_string or 'couchbase://127.0.0.1'}:{online_store_config.kv_port or '11210'}",
                ClusterOptions(
                    PasswordAuthenticator(
                        online_store_config.user or "Administrator",
                        online_store_config.password or "password",
                    ),
                    network="external",
                ),
            )

            self.bucket = self._cluster.bucket(
                online_store_config.bucket_name or "feast"
            )

        # Get the specific scope and collection
        scope = self.bucket.scope(scope_name)
        self.collection = scope.collection(collection_name)

        return self.collection

    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        """
        Write a batch of feature data to the online Couchbase store.

        Args:
            config: The RepoConfig for the current FeatureStore.
            table: Feast FeatureView.
            data: a list of quadruplets containing Feature data. Each
                  quadruplet contains an Entity Key, a dict containing feature
                  values, an event timestamp for the row, and
                  the created timestamp for the row if it exists.
            progress: Optional function to be called once every mini-batch of
                      rows is written to the online store. Can be used to
                      display progress.
        """
        warnings.warn(
            "This online store is an experimental feature in alpha development. "
            "Some functionality may still be unstable so functionality can change in the future.",
            RuntimeWarning,
        )
        project = config.project
        scope_name = f"{project}_{table.name}_scope"
        collection_name = f"{project}_{table.name}_collection"
        collection = self._get_conn(config, scope_name, collection_name)

        for entity_key, values, timestamp, created_ts in data:
            entity_key_str = serialize_entity_key(
                entity_key,
                entity_key_serialization_version=config.entity_key_serialization_version,
            ).hex()
            timestamp = _to_naive_utc(timestamp).isoformat()  # Convert to ISO format
            if created_ts is not None:
                created_ts = _to_naive_utc(
                    created_ts
                ).isoformat()  # Convert to ISO format

            for feature_name, val in values.items():
                document_id = _document_id(project, table, entity_key_str, feature_name)

                # Serialize the Protobuf to binary and then encode it in base64
                binary_value = val.SerializeToString()
                base64_value = base64.b64encode(binary_value).decode("utf-8")

                # Store metadata and base64-encoded Protobuf binary in JSON-compatible format
                document_content = {
                    "metadata": {
                        "event_ts": timestamp,
                        "created_ts": created_ts,
                        "feature_name": feature_name,
                    },
                    "value": base64_value,  # Store binary as base64 encoded string
                }

                try:
                    collection.upsert(
                        document_id, document_content
                    )  # Upsert the document
                except Exception as e:
                    logger.exception(f"Error upserting document {document_id}: {e}")

                if progress:
                    progress(1)

    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        """
        Read feature values pertaining to the requested entities from
        the online store.

        Args:
            config: The RepoConfig for the current FeatureStore.
            table: Feast FeatureView.
            entity_keys: a list of entity keys that should be read
                         from the FeatureStore.
            requested_features: Optional list of feature names to read.
        """
        warnings.warn(
            "This online store is an experimental feature in alpha development. "
            "Some functionality may still be unstable so functionality can change in the future.",
            RuntimeWarning,
        )
        project = config.project

        scope_name = f"{project}_{table.name}_scope"
        collection_name = f"{project}_{table.name}_collection"

        collection = self._get_conn(config, scope_name, collection_name)

        result: List[Tuple[Optional[datetime], Optional[Dict[str, Any]]]] = []
        for entity_key in entity_keys:
            entity_key_str = serialize_entity_key(
                entity_key,
                entity_key_serialization_version=config.entity_key_serialization_version,
            ).hex()
            try:
                features = {}
                for feature_name in requested_features or []:
                    document_id = _document_id(
                        project, table, entity_key_str, feature_name
                    )

                    # Fetch metadata and value (base64-encoded binary)
                    doc = collection.get(document_id)
                    content = doc.content_as[dict]  # Get the document content as a dict
                    event_ts_str = content["metadata"]["event_ts"]

                    # Convert event_ts from string (ISO format) to datetime object
                    event_ts = datetime.fromisoformat(event_ts_str)

                    base64_value = content["value"]

                    # Decode base64 back to Protobuf binary and then to ValueProto
                    binary_data = base64.b64decode(base64_value)
                    value = ValueProto()
                    value.ParseFromString(binary_data)  # Parse protobuf data

                    # Add the decoded value to the features dictionary
                    features[feature_name] = value

                result.append((event_ts, features))
            except DocumentNotFoundException:
                result.append((None, None))

        return result

    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[FeatureView],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ):
        """
        Update schema on DB, by creating and destroying tables accordingly.

        Args:
            config: The RepoConfig for the current FeatureStore.
            tables_to_delete: Tables to delete from the Online Store.
            tables_to_keep: Tables to keep in the Online Store.
            entities_to_delete: Entities to delete from the Online Store.
            entities_to_keep: Entities to keep in the Online Store.
            partial: Whether to partially update the schema.
        """
        warnings.warn(
            "This online store is an experimental feature in alpha development. "
            "Some functionality may still be unstable so functionality can change in the future.",
            RuntimeWarning,
        )
        project = config.project

        for table in tables_to_keep:
            scope_name = f"{project}_{table.name}_scope"
            collection_name = f"{project}_{table.name}_collection"
            self._get_conn(config, scope_name, collection_name)
            cm = self.bucket.collections()

            # Check and create scope
            try:
                cm.create_scope(scope_name)
                logger.info(f"Created scope: {scope_name}")
            except ScopeAlreadyExistsException:
                logger.error(f"Scope {scope_name} already exists")
            except Exception as e:
                logger.error(f"Error creating scope {scope_name}: {e}")

            # Check and create collection
            try:
                cm.create_collection(scope_name, collection_name)
                logger.info(
                    f"Created collection: {collection_name} in scope: {scope_name}"
                )
            except CollectionAlreadyExistsException:
                logger.error(
                    f"Collection {collection_name} already exists in {scope_name}"
                )
            except Exception as e:
                logger.error(f"Error creating collection {collection_name}: {e}")

    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ):
        """
        Delete tables from the database.

        Args:
            config: The RepoConfig for the current FeatureStore.
            tables: Tables to delete from the feature repo.
            entities: Entities to delete from the feature repo.
        """
        warnings.warn(
            "This online store is an experimental feature in alpha development. "
            "Some functionality may still be unstable so functionality can change in the future.",
            RuntimeWarning,
        )
        project = config.project

        for table in tables:
            scope_name = f"{project}_{table.name}_scope"
            collection_name = f"{project}_{table.name}_collection"
            self._get_conn(config, scope_name, collection_name)
            cm = self.bucket.collections()
            try:
                # dropping the scope will also drop the nested collection(s)
                cm.drop_scope(scope_name)
            except Exception as e:
                logger.error(f"Error removing collection or scope: {e}")


def _document_id(
    project: str, table: FeatureView, entity_key_str: str, feature_name: str
) -> str:
    return f"{project}:{table.name}:{entity_key_str}:{feature_name}"


def _to_naive_utc(ts: datetime):
    if ts.tzinfo is None:
        return ts
    else:
        return ts.astimezone(pytz.utc).replace(tzinfo=None)



from abc import ABC, abstractmethod
from typing import List, Optional

from feast.infra.compute_engines.dag.context import ExecutionContext
from feast.infra.compute_engines.dag.value import DAGValue


class DAGNode(ABC):
    name: str
    inputs: List["DAGNode"]
    outputs: List["DAGNode"]

    def __init__(self, name: str, inputs: Optional[List["DAGNode"]] = None):
        self.name = name
        self.inputs: List["DAGNode"] = []
        self.outputs: List["DAGNode"] = []

        for node in inputs or []:
            self.add_input(node)

    def add_input(self, node: "DAGNode"):
        if node in self.inputs:
            raise ValueError(f"Input node {node.name} already added to {self.name}")
        self.inputs.append(node)
        node.outputs.append(self)

    def get_input_values(self, context: ExecutionContext) -> List[DAGValue]:
        input_values = []
        for input_node in self.inputs:
            if input_node.name not in context.node_outputs:
                raise KeyError(
                    f"Missing output for input node '{input_node.name}' in context."
                )
            input_values.append(context.node_outputs[input_node.name])
        return input_values

    def get_single_input_value(self, context: ExecutionContext) -> DAGValue:
        if len(self.inputs) != 1:
            raise RuntimeError(
                f"DAGNode '{self.name}' expected exactly 1 input, but got {len(self.inputs)}."
            )
        input_node = self.inputs[0]
        if input_node.name not in context.node_outputs:
            raise KeyError(
                f"Missing output for input node '{input_node.name}' in context."
            )
        return context.node_outputs[input_node.name]

    @abstractmethod
    def execute(self, context: ExecutionContext) -> DAGValue: ...



import inspect
from typing import Any, Callable, Optional, cast, get_type_hints

import dill
import pandas as pd
import pyarrow

from feast.field import Field, from_value_type
from feast.protos.feast.core.Transformation_pb2 import (
    UserDefinedFunctionV2 as UserDefinedFunctionProto,
)
from feast.transformation.base import Transformation
from feast.transformation.mode import TransformationMode
from feast.type_map import (
    python_type_to_feast_value_type,
)


class PandasTransformation(Transformation):
    def __new__(
        cls,
        udf: Optional[Callable[[Any], Any]] = None,
        udf_string: Optional[str] = None,
        name: Optional[str] = None,
        tags: Optional[dict[str, str]] = None,
        description: str = "",
        owner: str = "",
    ) -> "PandasTransformation":
        # Handle Ray deserialization where parameters may not be provided
        if udf is None and udf_string is None:
            # Create a bare instance for deserialization
            instance = object.__new__(cls)
            return cast("PandasTransformation", instance)

        # Ensure required parameters are not None before calling parent constructor
        if udf is None:
            raise ValueError("udf parameter cannot be None")
        if udf_string is None:
            raise ValueError("udf_string parameter cannot be None")

        return cast(
            "PandasTransformation",
            super(PandasTransformation, cls).__new__(
                cls,
                mode=TransformationMode.PANDAS,
                udf=udf,
                name=name,
                udf_string=udf_string,
                tags=tags,
                description=description,
                owner=owner,
            ),
        )

    def __init__(
        self,
        udf: Optional[Callable[[Any], Any]] = None,
        udf_string: Optional[str] = None,
        name: Optional[str] = None,
        tags: Optional[dict[str, str]] = None,
        description: str = "",
        owner: str = "",
        *args,
        **kwargs,
    ):
        # Handle Ray deserialization where parameters may not be provided
        if udf is None and udf_string is None:
            # Early return for deserialization - don't initialize
            return

        # Ensure required parameters are not None before calling parent constructor
        if udf is None:
            raise ValueError("udf parameter cannot be None")
        if udf_string is None:
            raise ValueError("udf_string parameter cannot be None")

        return_annotation = get_type_hints(udf).get("return", inspect._empty)
        if return_annotation not in (inspect._empty, pd.DataFrame):
            raise TypeError(
                f"return signature for PandasTransformation should be pd.DataFrame, instead got {return_annotation}"
            )

        super().__init__(
            mode=TransformationMode.PANDAS,
            udf=udf,
            name=name,
            udf_string=udf_string,
            tags=tags,
            description=description,
            owner=owner,
        )

    def transform_arrow(
        self, pa_table: pyarrow.Table, features: list[Field]
    ) -> pyarrow.Table:
        output_df_pandas = self.udf(pa_table.to_pandas())
        return pyarrow.Table.from_pandas(output_df_pandas)

    def transform(self, inputs: pd.DataFrame) -> pd.DataFrame:
        return self.udf(inputs)

    def infer_features(
        self,
        random_input: dict[str, list[Any]],
        *args,
        **kwargs,
    ) -> list[Field]:
        df = pd.DataFrame.from_dict(random_input)
        output_df: pd.DataFrame = self.transform(df)

        fields = []
        for feature_name, feature_type in zip(output_df.columns, output_df.dtypes):
            feature_value = output_df[feature_name].tolist()
            if len(feature_value) <= 0:
                raise TypeError(
                    f"Failed to infer type for feature '{feature_name}' with value "
                    + f"'{feature_value}' since no items were returned by the UDF."
                )
            fields.append(
                Field(
                    name=feature_name,
                    dtype=from_value_type(
                        python_type_to_feast_value_type(
                            feature_name,
                            value=feature_value[0],
                            type_name=str(feature_type),
                        )
                    ),
                )
            )
        return fields

    def __eq__(self, other):
        if not isinstance(other, PandasTransformation):
            raise TypeError(
                "Comparisons should only involve PandasTransformation class objects."
            )

        if (
            self.udf_string != other.udf_string
            or self.udf.__code__.co_code != other.udf.__code__.co_code
        ):
            return False

        return True

    @classmethod
    def from_proto(cls, user_defined_function_proto: UserDefinedFunctionProto):
        return PandasTransformation(
            udf=dill.loads(user_defined_function_proto.body),
            udf_string=user_defined_function_proto.body_text,
        )



from datetime import datetime, timezone
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
)

from google.protobuf.timestamp_pb2 import Timestamp
from ikvpy.client import IKVReader, IKVWriter
from ikvpy.clientoptions import ClientOptions, ClientOptionsBuilder
from ikvpy.document import IKVDocument, IKVDocumentBuilder
from ikvpy.factory import create_new_reader, create_new_writer
from pydantic import StrictStr

from feast import Entity, FeatureView, utils
from feast.infra.online_stores.helpers import compute_entity_id
from feast.infra.online_stores.online_store import OnlineStore
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import FeastConfigBaseModel, RepoConfig

PRIMARY_KEY_FIELD_NAME: str = "_entity_key"
EVENT_CREATION_TIMESTAMP_FIELD_NAME: str = "_event_timestamp"
CREATION_TIMESTAMP_FIELD_NAME: str = "_created_timestamp"


class IKVOnlineStoreConfig(FeastConfigBaseModel):
    """Online store config for IKV store"""

    type: Literal["ikv"] = "ikv"
    """Online store type selector"""

    account_id: StrictStr
    """(Required) IKV account id"""

    account_passkey: StrictStr
    """(Required) IKV account passkey"""

    store_name: StrictStr
    """(Required) IKV store name"""

    mount_directory: Optional[StrictStr] = None
    """(Required only for reader) IKV mount point i.e. directory for storing IKV data locally."""


class IKVOnlineStore(OnlineStore):
    """
    IKV (inlined.io key value) store implementation of the online store interface.
    """

    # lazy initialization
    _reader: Optional[IKVReader] = None
    _writer: Optional[IKVWriter] = None

    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        """
        Writes a batch of feature rows to the online store.

        If a tz-naive timestamp is passed to this method, it is assumed to be UTC.

        Args:
            config: The config for the current feature store.
            table: Feature view to which these feature rows correspond.
            data: A list of quadruplets containing feature data. Each quadruplet contains an entity
                key, a dict containing feature values, an event timestamp for the row, and the created
                timestamp for the row if it exists.
            progress: Function to be called once a batch of rows is written to the online store, used
                to show progress.
        """
        self._init_writer(config=config)
        assert self._writer is not None

        for entity_key, features, event_timestamp, _ in data:
            entity_id: str = compute_entity_id(
                entity_key,
                entity_key_serialization_version=config.entity_key_serialization_version,
            )
            document: IKVDocument = IKVOnlineStore._create_document(
                entity_id, table, features, event_timestamp
            )
            self._writer.upsert_fields(document)
            if progress:
                progress(1)

    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        """
        Reads features values for the given entity keys.

        Args:
            config: The config for the current feature store.
            table: The feature view whose feature values should be read.
            entity_keys: The list of entity keys for which feature values should be read.
            requested_features: The list of features that should be read.

        Returns:
            A list of the same length as entity_keys. Each item in the list is a tuple where the first
            item is the event timestamp for the row, and the second item is a dict mapping feature names
            to values, which are returned in proto format.
        """
        self._init_reader(config=config)

        if not len(entity_keys):
            return []

        # create IKV primary keys
        primary_keys = [
            compute_entity_id(ek, config.entity_key_serialization_version)
            for ek in entity_keys
        ]

        # create IKV field names
        if requested_features is None:
            requested_features = []

        field_names: List[Optional[str]] = [None] * (1 + len(requested_features))
        field_names[0] = EVENT_CREATION_TIMESTAMP_FIELD_NAME
        for i, fn in enumerate(requested_features):
            field_names[i + 1] = IKVOnlineStore._create_ikv_field_name(table, fn)

        assert self._reader is not None
        value_iter = self._reader.multiget_bytes_values(
            bytes_primary_keys=[],
            str_primary_keys=primary_keys,
            field_names=field_names,
        )

        # decode results
        return [
            IKVOnlineStore._decode_fields_for_primary_key(
                requested_features, value_iter
            )
            for _ in range(0, len(primary_keys))
        ]

    @staticmethod
    def _decode_fields_for_primary_key(
        requested_features: List[str], value_iter: Iterator[Optional[bytes]]
    ) -> Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]:
        # decode timestamp
        dt: Optional[datetime] = None
        dt_bytes = next(value_iter)
        if dt_bytes:
            proto_timestamp = Timestamp()
            proto_timestamp.ParseFromString(dt_bytes)
            dt = datetime.fromtimestamp(proto_timestamp.seconds, tz=timezone.utc)

        # decode other features
        features = {}
        for requested_feature in requested_features:
            value_proto_bytes: Optional[bytes] = next(value_iter)
            if value_proto_bytes:
                value_proto = ValueProto()
                value_proto.ParseFromString(value_proto_bytes)
                features[requested_feature] = value_proto

        return dt, features

    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[FeatureView],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ):
        """
        Reconciles cloud resources with the specified set of Feast objects.

        Args:
            config: The config for the current feature store.
            tables_to_delete: Feature views whose corresponding infrastructure should be deleted.
            tables_to_keep: Feature views whose corresponding infrastructure should not be deleted, and
                may need to be updated.
            entities_to_delete: Entities whose corresponding infrastructure should be deleted.
            entities_to_keep: Entities whose corresponding infrastructure should not be deleted, and
                may need to be updated.
            partial: If true, tables_to_delete and tables_to_keep are not exhaustive lists, so
                infrastructure corresponding to other feature views should be not be touched.
        """
        self._init_writer(config=config)
        assert self._writer is not None

        # note: we assume tables_to_keep does not overlap with tables_to_delete

        for feature_view in tables_to_delete:
            # each field in an IKV document is prefixed by the feature-view's name
            self._writer.drop_fields_by_name_prefix([feature_view.name])

    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ):
        """
        Tears down all cloud resources for the specified set of Feast objects.

        Args:
            config: The config for the current feature store.
            tables: Feature views whose corresponding infrastructure should be deleted.
            entities: Entities whose corresponding infrastructure should be deleted.
        """
        self._init_writer(config=config)
        assert self._writer is not None

        # drop fields corresponding to this feature-view
        for feature_view in tables:
            self._writer.drop_fields_by_name_prefix([feature_view.name])

        # shutdown clients
        self._writer.shutdown()
        self._writer = None

        if self._reader is not None:
            self._reader.shutdown()
            self._reader = None

    @staticmethod
    def _create_ikv_field_name(feature_view: FeatureView, feature_name: str) -> str:
        return "{}_{}".format(feature_view.name, feature_name)

    @staticmethod
    def _create_document(
        entity_id: str,
        feature_view: FeatureView,
        values: Dict[str, ValueProto],
        event_timestamp: datetime,
    ) -> IKVDocument:
        """Converts feast key-value pairs into an IKV document."""

        # initialie builder by inserting primary key and row creation timestamp
        event_timestamp_seconds = int(utils.make_tzaware(event_timestamp).timestamp())
        event_timestamp_seconds_proto = Timestamp()
        event_timestamp_seconds_proto.seconds = event_timestamp_seconds

        # event_timestamp_str: str = utils.make_tzaware(event_timestamp).isoformat()
        builder = (
            IKVDocumentBuilder()
            .put_string_field(PRIMARY_KEY_FIELD_NAME, entity_id)
            .put_bytes_field(
                EVENT_CREATION_TIMESTAMP_FIELD_NAME,
                event_timestamp_seconds_proto.SerializeToString(),
            )
        )

        for feature_name, feature_value in values.items():
            field_name = IKVOnlineStore._create_ikv_field_name(
                feature_view, feature_name
            )
            builder.put_bytes_field(field_name, feature_value.SerializeToString())

        return builder.build()

    def _init_writer(self, config: RepoConfig):
        """Initializes ikv writer client."""
        # initialize writer
        if self._writer is None:
            online_config = config.online_store
            assert isinstance(online_config, IKVOnlineStoreConfig)
            client_options = IKVOnlineStore._config_to_client_options(online_config)

            self._writer = create_new_writer(client_options)
            self._writer.startup()  # blocking operation

    def _init_reader(self, config: RepoConfig):
        """Initializes ikv reader client."""
        # initialize reader
        if self._reader is None:
            online_config = config.online_store
            assert isinstance(online_config, IKVOnlineStoreConfig)
            client_options = IKVOnlineStore._config_to_client_options(online_config)

            if online_config.mount_directory and len(online_config.mount_directory) > 0:
                self._reader = create_new_reader(client_options)
                self._reader.startup()  # blocking operation

    @staticmethod
    def _config_to_client_options(config: IKVOnlineStoreConfig) -> ClientOptions:
        """Utility for IKVOnlineStoreConfig to IKV ClientOptions conversion."""
        builder = (
            ClientOptionsBuilder()
            .with_account_id(config.account_id)
            .with_account_passkey(config.account_passkey)
            .with_store_name(config.store_name)
        )

        if config.mount_directory and len(config.mount_directory) > 0:
            builder = builder.with_mount_directory(config.mount_directory)

        return builder.build()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

Copyright 2020 The Feast Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import feast.core.DataSource_pb2
import feast.core.Feature_pb2
import google.protobuf.descriptor
import google.protobuf.duration_pb2
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.timestamp_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class FeatureView(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SPEC_FIELD_NUMBER: builtins.int
    META_FIELD_NUMBER: builtins.int
    @property
    def spec(self) -> global___FeatureViewSpec:
        """User-specified specifications of this feature view."""
    @property
    def meta(self) -> global___FeatureViewMeta:
        """System-populated metadata for this feature view."""
    def __init__(
        self,
        *,
        spec: global___FeatureViewSpec | None = ...,
        meta: global___FeatureViewMeta | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["meta", b"meta", "spec", b"spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["meta", b"meta", "spec", b"spec"]) -> None: ...

global___FeatureView = FeatureView

class FeatureViewSpec(google.protobuf.message.Message):
    """Next available id: 13
    TODO(adchia): refactor common fields from this and ODFV into separate metadata proto
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class TagsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    PROJECT_FIELD_NUMBER: builtins.int
    ENTITIES_FIELD_NUMBER: builtins.int
    FEATURES_FIELD_NUMBER: builtins.int
    ENTITY_COLUMNS_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    TAGS_FIELD_NUMBER: builtins.int
    OWNER_FIELD_NUMBER: builtins.int
    TTL_FIELD_NUMBER: builtins.int
    BATCH_SOURCE_FIELD_NUMBER: builtins.int
    STREAM_SOURCE_FIELD_NUMBER: builtins.int
    ONLINE_FIELD_NUMBER: builtins.int
    OFFLINE_FIELD_NUMBER: builtins.int
    SOURCE_VIEWS_FIELD_NUMBER: builtins.int
    name: builtins.str
    """Name of the feature view. Must be unique. Not updated."""
    project: builtins.str
    """Name of Feast project that this feature view belongs to."""
    @property
    def entities(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """List of names of entities associated with this feature view."""
    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.Feature_pb2.FeatureSpecV2]:
        """List of specifications for each feature defined as part of this feature view."""
    @property
    def entity_columns(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.Feature_pb2.FeatureSpecV2]:
        """List of specifications for each entity defined as part of this feature view."""
    description: builtins.str
    """Description of the feature view."""
    @property
    def tags(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """User defined metadata"""
    owner: builtins.str
    """Owner of the feature view."""
    @property
    def ttl(self) -> google.protobuf.duration_pb2.Duration:
        """Features in this feature view can only be retrieved from online serving
        younger than ttl. Ttl is measured as the duration of time between
        the feature's event timestamp and when the feature is retrieved
        Feature values outside ttl will be returned as unset values and indicated to end user
        """
    @property
    def batch_source(self) -> feast.core.DataSource_pb2.DataSource:
        """Batch/Offline DataSource where this view can retrieve offline feature data."""
    @property
    def stream_source(self) -> feast.core.DataSource_pb2.DataSource:
        """Streaming DataSource from where this view can consume "online" feature data."""
    online: builtins.bool
    """Whether these features should be served online or not
    This is also used to determine whether the features should be written to the online store
    """
    offline: builtins.bool
    """Whether these features should be written to the offline store"""
    @property
    def source_views(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FeatureViewSpec]: ...
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        project: builtins.str = ...,
        entities: collections.abc.Iterable[builtins.str] | None = ...,
        features: collections.abc.Iterable[feast.core.Feature_pb2.FeatureSpecV2] | None = ...,
        entity_columns: collections.abc.Iterable[feast.core.Feature_pb2.FeatureSpecV2] | None = ...,
        description: builtins.str = ...,
        tags: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        owner: builtins.str = ...,
        ttl: google.protobuf.duration_pb2.Duration | None = ...,
        batch_source: feast.core.DataSource_pb2.DataSource | None = ...,
        stream_source: feast.core.DataSource_pb2.DataSource | None = ...,
        online: builtins.bool = ...,
        offline: builtins.bool = ...,
        source_views: collections.abc.Iterable[global___FeatureViewSpec] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["batch_source", b"batch_source", "stream_source", b"stream_source", "ttl", b"ttl"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["batch_source", b"batch_source", "description", b"description", "entities", b"entities", "entity_columns", b"entity_columns", "features", b"features", "name", b"name", "offline", b"offline", "online", b"online", "owner", b"owner", "project", b"project", "source_views", b"source_views", "stream_source", b"stream_source", "tags", b"tags", "ttl", b"ttl"]) -> None: ...

global___FeatureViewSpec = FeatureViewSpec

class FeatureViewMeta(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREATED_TIMESTAMP_FIELD_NUMBER: builtins.int
    LAST_UPDATED_TIMESTAMP_FIELD_NUMBER: builtins.int
    MATERIALIZATION_INTERVALS_FIELD_NUMBER: builtins.int
    @property
    def created_timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """Time where this Feature View is created"""
    @property
    def last_updated_timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """Time where this Feature View is last updated"""
    @property
    def materialization_intervals(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___MaterializationInterval]:
        """List of pairs (start_time, end_time) for which this feature view has been materialized."""
    def __init__(
        self,
        *,
        created_timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        last_updated_timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        materialization_intervals: collections.abc.Iterable[global___MaterializationInterval] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["created_timestamp", b"created_timestamp", "last_updated_timestamp", b"last_updated_timestamp"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["created_timestamp", b"created_timestamp", "last_updated_timestamp", b"last_updated_timestamp", "materialization_intervals", b"materialization_intervals"]) -> None: ...

global___FeatureViewMeta = FeatureViewMeta

class MaterializationInterval(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    START_TIME_FIELD_NUMBER: builtins.int
    END_TIME_FIELD_NUMBER: builtins.int
    @property
    def start_time(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
    @property
    def end_time(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
    def __init__(
        self,
        *,
        start_time: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        end_time: google.protobuf.timestamp_pb2.Timestamp | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["end_time", b"end_time", "start_time", b"start_time"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["end_time", b"end_time", "start_time", b"start_time"]) -> None: ...

global___MaterializationInterval = MaterializationInterval

class FeatureViewList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FEATUREVIEWS_FIELD_NUMBER: builtins.int
    @property
    def featureviews(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FeatureView]: ...
    def __init__(
        self,
        *,
        featureviews: collections.abc.Iterable[global___FeatureView] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["featureviews", b"featureviews"]) -> None: ...

global___FeatureViewList = FeatureViewList



import logging
import os

import jwt

from feast.permissions.auth_model import KubernetesAuthConfig
from feast.permissions.client.auth_client_manager import AuthenticationClientManager

logger = logging.getLogger(__name__)


class KubernetesAuthClientManager(AuthenticationClientManager):
    def __init__(self, auth_config: KubernetesAuthConfig):
        self.auth_config = auth_config
        self.token_file_path = "/var/run/secrets/kubernetes.io/serviceaccount/token"

    def get_token(self):
        intra_communication_base64 = os.getenv("INTRA_COMMUNICATION_BASE64")
        # If intra server communication call
        if intra_communication_base64:
            payload = {
                "sub": f":::{intra_communication_base64}",  # Subject claim
            }

            return jwt.encode(payload, "")

        try:
            token = self._read_token_from_file()
            return token
        except Exception as e:
            logger.info(f"Error reading token from file: {e}")
            logger.info("Attempting to read token from environment variable.")
            try:
                token = self._read_token_from_env()
                return token
            except Exception as env_e:
                logger.exception(
                    f"Error reading token from environment variable: {env_e}"
                )
                raise env_e

    def _read_token_from_file(self):
        try:
            with open(self.token_file_path, "r") as file:
                token = file.read().strip()
            return token
        except Exception as e:
            raise e

    def _read_token_from_env(self):
        token = os.getenv("LOCAL_K8S_TOKEN")
        if not token:
            raise KeyError("LOCAL_K8S_TOKEN environment variable is not set.")
        return token



from typing import List

from feast.infra.compute_engines.dag.context import ExecutionContext
from feast.infra.compute_engines.dag.node import DAGNode
from feast.infra.compute_engines.dag.value import DAGValue


class ExecutionPlan:
    """
    ExecutionPlan represents an ordered sequence of DAGNodes that together define
    a data processing pipeline for feature materialization or historical retrieval.

    This plan is constructed as a topological sort of the DAG — meaning that each
    node appears after all its input dependencies. The plan is executed in order,
    caching intermediate results (`DAGValue`) so that each node can reuse outputs
    from upstream nodes without recomputation.

    Key Concepts:
    - DAGNode: Each node performs a specific logical step (e.g., read, aggregate, join).
    - DAGValue: Output of a node, includes data (e.g., Spark DataFrame) and metadata.
    - ExecutionContext: Contains runtime information (config, registry, stores, entity_df).
    - node_outputs: A cache of intermediate results keyed by node name.

    Usage:
        plan = ExecutionPlan(dag_nodes)
        result = plan.execute(context)

    This design enables modular compute backends (e.g., Spark, Pandas, Arrow), where
    each node defines its execution logic independently while benefiting from shared
    execution orchestration, caching, and context injection.

    Example:
        DAG:
            ReadNode -> TransformNode -> AggregateNode -> -> WriteNode
                     -> JoinNode      ->

        Execution proceeds step by step, passing intermediate DAGValues through
        the plan while respecting node dependencies and formats.

    This approach is inspired by execution DAGs in systems like Apache Spark,
    Apache Beam, and Dask — but specialized for Feast’s feature computation domain.
    """

    def __init__(self, nodes: List[DAGNode]):
        self.nodes = nodes

    def execute(self, context: ExecutionContext) -> DAGValue:
        context.node_outputs = {}

        for node in self.nodes:
            output = node.execute(context)
            context.node_outputs[node.name] = output

        return context.node_outputs[self.nodes[-1].name]

    def to_sql(self, context: ExecutionContext) -> str:
        """
        Generate SQL query for the entire execution plan.
        This is a placeholder and should be implemented in subclasses.
        """
        raise NotImplementedError("SQL generation is not implemented yet.")

    def to_dag(self) -> str:
        """
        Render the DAG as a multiline string with full node expansion (no visited shortcut).
        """

        def walk(node: DAGNode, indent: int = 0) -> List[str]:
            prefix = "  " * indent
            lines = [f"{prefix}- {node.name}"]
            for input_node in node.inputs:
                lines.extend(walk(input_node, indent + 1))
            return lines

        return "\n".join(walk(self.nodes[-1]))



from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import pandas as pd
import pyarrow

from feast import FeatureView, RepoConfig
from feast.data_source import _DATA_SOURCE_FOR_OFFLINE_STORE, DataSource
from feast.errors import FeastOfflineStoreInvalidName
from feast.feature_logging import LoggingConfig, LoggingSource
from feast.infra.offline_stores.offline_store import OfflineStore, RetrievalJob
from feast.infra.offline_stores.offline_utils import get_offline_store_from_config
from feast.infra.registry.base_registry import BaseRegistry
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.repo_config import (
    FeastConfigBaseModel,
    get_offline_config_from_type,
    get_offline_store_type,
)


class HybridOfflineStoreConfig(FeastConfigBaseModel):
    type: str = "hybrid_offline_store.HybridOfflineStore"

    class OfflineStoresWithConfig(FeastConfigBaseModel):
        type: str
        conf: Dict[str, Any]

    offline_stores: Optional[List[OfflineStoresWithConfig]]


class HybridOfflineStore(OfflineStore):
    _instance: Optional["HybridOfflineStore"] = None
    _initialized: bool
    offline_stores: Dict[str, OfflineStore]

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(HybridOfflineStore, cls).__new__(cls)
            cls._instance._initialized = False
            cls._instance.offline_stores = {}
        return cls._instance

    def _initialize_offline_stores(self, config: RepoConfig):
        if self._initialized:
            return
        for store_cfg in getattr(config.offline_store, "offline_stores", []):
            try:
                offline_store_type = get_offline_store_type(store_cfg.type)
                config_cls = get_offline_config_from_type(store_cfg.type)
                config_instance = config_cls(**store_cfg.conf)
                store = get_offline_store_from_config(config_instance)
                self.offline_stores[offline_store_type] = store
            except FeastOfflineStoreInvalidName as e:
                raise FeastOfflineStoreInvalidName(
                    f"Failed to initialize Hybrid offline store {store_cfg.type}: {e}"
                )
        self._initialized = True

    def get_source_key_from_type(
        self, source_type: DataSourceProto.SourceType.ValueType
    ) -> Optional[str]:
        if source_type not in list(_DATA_SOURCE_FOR_OFFLINE_STORE.keys()):
            raise ValueError(
                f"Unsupported DataSource type for HybridOfflineStore: {source_type}."
                f"Supported types are: {list(_DATA_SOURCE_FOR_OFFLINE_STORE.keys())}"
            )
        return _DATA_SOURCE_FOR_OFFLINE_STORE.get(source_type, None)

    def _get_offline_store_for_feature_view(
        self, feature_view: FeatureView, config: RepoConfig
    ) -> OfflineStore:
        self._initialize_offline_stores(config)
        source_type = feature_view.batch_source.source_type()
        store_key = self.get_source_key_from_type(source_type)
        if store_key is None:
            raise ValueError(
                f"Unsupported FeatureView batch_source type: {source_type}"
            )
        return self.offline_stores[store_key]

    def _get_offline_store_for_source(
        self, data_source: DataSource, config: RepoConfig
    ) -> OfflineStore:
        self._initialize_offline_stores(config)
        source_type = data_source.source_type()
        store_key = self.get_source_key_from_type(source_type)
        if store_key is None:
            raise ValueError(f"Unsupported DataSource type: {source_type}")
        return self.offline_stores[store_key]

    @staticmethod
    def get_historical_features(
        config: RepoConfig,
        feature_views: List[FeatureView],
        feature_refs: List[str],
        entity_df: Union[pd.DataFrame, str],
        registry: BaseRegistry,
        project: str,
        full_feature_names: bool = False,
    ) -> RetrievalJob:
        # TODO: Multiple data sources can be supported when feature store use compute engine
        # for getting historical features
        data_source = None
        for feature_view in feature_views:
            if not feature_view.batch_source:
                raise ValueError(
                    "HybridOfflineStore only supports feature views with DataSource as source. "
                )
            if not data_source:
                data_source = feature_view.batch_source
            elif data_source != feature_view.batch_source:
                raise ValueError(
                    "All feature views must have the same batch source for HybridOfflineStore."
                )

        store = HybridOfflineStore()._get_offline_store_for_feature_view(
            feature_views[0], config
        )
        return store.get_historical_features(
            config,
            feature_views,
            feature_refs,
            entity_df,
            registry,
            project,
            full_feature_names,
        )

    @staticmethod
    def pull_latest_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        start_date: datetime,
        end_date: datetime,
    ) -> RetrievalJob:
        store = HybridOfflineStore()._get_offline_store_for_source(data_source, config)
        return store.pull_latest_from_table_or_query(
            config,
            data_source,
            join_key_columns,
            feature_name_columns,
            timestamp_field,
            created_timestamp_column,
            start_date,
            end_date,
        )

    @staticmethod
    def pull_all_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
    ) -> RetrievalJob:
        store = HybridOfflineStore()._get_offline_store_for_source(data_source, config)
        return store.pull_all_from_table_or_query(
            config,
            data_source,
            join_key_columns,
            feature_name_columns,
            timestamp_field,
            created_timestamp_column,
            start_date,
            end_date,
        )

    @staticmethod
    def write_logged_features(
        config: RepoConfig,
        data: Union[pyarrow.Table, Path],
        source: LoggingSource,
        logging_config: LoggingConfig,
        registry: BaseRegistry,
    ):
        raise NotImplementedError(
            "HybridOfflineStore does not support write_logged_features. "
            "Please use the specific offline store for logging."
        )

    @staticmethod
    def offline_write_batch(
        config: RepoConfig,
        feature_view: FeatureView,
        table: pyarrow.Table,
        progress: Optional[Callable[[int], Any]],
    ):
        store = HybridOfflineStore()._get_offline_store_for_feature_view(
            feature_view, config
        )
        return store.offline_write_batch(config, feature_view, table, progress)

    def validate_data_source(
        self,
        config: RepoConfig,
        data_source: DataSource,
    ):
        store = self._get_offline_store_for_source(data_source, config)
        return store.validate_data_source(config, data_source)

    def get_table_column_names_and_types_from_data_source(
        self,
        config: RepoConfig,
        data_source: DataSource,
    ) -> Iterable[Tuple[str, str]]:
        store = self._get_offline_store_for_source(data_source, config)
        return store.get_table_column_names_and_types_from_data_source(
            config, data_source
        )



from dataclasses import dataclass

import dill

from feast import FeatureView
from feast.infra.passthrough_provider import PassthroughProvider
from feast.protos.feast.core.FeatureView_pb2 import FeatureView as FeatureViewProto


@dataclass
class SerializedArtifacts:
    """Class to assist with serializing unpicklable artifacts to be passed to the compute engine."""

    feature_view_proto: str
    repo_config_byte: str

    @classmethod
    def serialize(cls, feature_view, repo_config):
        # serialize to proto
        feature_view_proto = feature_view.to_proto().SerializeToString()

        # serialize repo_config to disk. Will be used to instantiate the online store
        repo_config_byte = dill.dumps(repo_config)

        return SerializedArtifacts(
            feature_view_proto=feature_view_proto, repo_config_byte=repo_config_byte
        )

    def unserialize(self):
        # unserialize
        proto = FeatureViewProto()
        proto.ParseFromString(self.feature_view_proto)
        feature_view = FeatureView.from_proto(proto)

        # load
        repo_config = dill.loads(self.repo_config_byte)

        provider = PassthroughProvider(repo_config)
        online_store = provider.online_store
        offline_store = provider.offline_store
        return feature_view, online_store, offline_store, repo_config
